"""Utils for auto classification estimator"""
# pylint: disable=bad-whitespace,missing-function-docstring
import os
import math

import mxnet as mx
from mxnet import nd
from mxnet import gluon
from mxnet.gluon.data.vision import transforms
from ....data import imagenet

def rec_batch_fn(batch, ctx):
    data = gluon.utils.split_and_load(batch.data[0], ctx_list=ctx, batch_axis=0, even_split=False)
    label = gluon.utils.split_and_load(batch.label[0], ctx_list=ctx, batch_axis=0, even_split=False)
    return data, label

def get_data_rec(rec_train, rec_train_idx, rec_val, rec_val_idx, batch_size, num_workers, input_size, crop_ratio):
    rec_train = os.path.expanduser(rec_train)
    rec_train_idx = os.path.expanduser(rec_train_idx)
    rec_val = os.path.expanduser(rec_val)
    rec_val_idx = os.path.expanduser(rec_val_idx)
    jitter_param = 0.4
    lighting_param = 0.1
    input_size = input_size
    crop_ratio = crop_ratio if crop_ratio > 0 else 0.875
    resize = int(math.ceil(input_size / crop_ratio))
    mean_rgb = [123.68, 116.779, 103.939]
    std_rgb = [58.393, 57.12, 57.375]

    train_data = mx.io.ImageRecordIter(
        path_imgrec         = rec_train,
        path_imgidx         = rec_train_idx,
        preprocess_threads  = num_workers,
        shuffle             = True,
        batch_size          = batch_size,

        data_shape          = (3, input_size, input_size),
        mean_r              = mean_rgb[0],
        mean_g              = mean_rgb[1],
        mean_b              = mean_rgb[2],
        std_r               = std_rgb[0],
        std_g               = std_rgb[1],
        std_b               = std_rgb[2],
        rand_mirror         = True,
        random_resized_crop = True,
        max_aspect_ratio    = 4. / 3.,
        min_aspect_ratio    = 3. / 4.,
        max_random_area     = 1,
        min_random_area     = 0.08,
        brightness          = jitter_param,
        saturation          = jitter_param,
        contrast            = jitter_param,
        pca_noise           = lighting_param,
    )
    val_data = mx.io.ImageRecordIter(
        path_imgrec         = rec_val,
        path_imgidx         = rec_val_idx,
        preprocess_threads  = num_workers,
        shuffle             = False,
        batch_size          = batch_size,

        resize              = resize,
        data_shape          = (3, input_size, input_size),
        mean_r              = mean_rgb[0],
        mean_g              = mean_rgb[1],
        mean_b              = mean_rgb[2],
        std_r               = std_rgb[0],
        std_g               = std_rgb[1],
        std_b               = std_rgb[2],
    )
    return train_data, val_data, rec_batch_fn

def loader_batch_fn(batch, ctx):
    if batch[0].shape[0] < len(ctx):
        # if # sample is less than # ctx, reduce the # ctx
        ctx = ctx[:batch[0].shape[0]]
    data = gluon.utils.split_and_load(batch[0], ctx_list=ctx, batch_axis=0, even_split=False)
    label = gluon.utils.split_and_load(batch[1], ctx_list=ctx, batch_axis=0, even_split=False)
    return data, label

def get_data_loader(data_dir, batch_size, num_workers, input_size, crop_ratio, train_dataset=None, val_dataset=None):
    normalize = transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    jitter_param = 0.4
    lighting_param = 0.1
    crop_ratio = crop_ratio if crop_ratio > 0 else 0.875
    resize = int(math.ceil(input_size / crop_ratio))

    transform_train = transforms.Compose([
        transforms.RandomResizedCrop(input_size),
        transforms.RandomFlipLeftRight(),
        transforms.RandomColorJitter(brightness=jitter_param, contrast=jitter_param,
                                     saturation=jitter_param),
        transforms.RandomLighting(lighting_param),
        transforms.ToTensor(),
        normalize
    ])
    transform_test = transforms.Compose([
        transforms.Resize(resize, keep_ratio=True),
        transforms.CenterCrop(input_size),
        transforms.ToTensor(),
        normalize
    ])

    if not train_dataset:
        train_dataset = imagenet.classification.ImageNet(data_dir, train=True)
    if not val_dataset:
        val_dataset = imagenet.classification.ImageNet(data_dir, train=False)

    train_data = gluon.data.DataLoader(
        train_dataset.transform_first(transform_train),
        batch_size=batch_size, shuffle=True, last_batch='discard', num_workers=num_workers)
    val_data = gluon.data.DataLoader(
        val_dataset.transform_first(transform_test),
        batch_size=batch_size, shuffle=False, num_workers=num_workers)

    return train_data, val_data, loader_batch_fn

def mixup_transform(label, classes, lam=1, eta=0.0):
    if isinstance(label, nd.NDArray):
        label = [label]
    res = []
    for l in label:
        y1 = l.one_hot(classes, on_value=1 - eta + eta/classes, off_value=eta/classes)
        y2 = l[::-1].one_hot(classes, on_value=1 - eta + eta/classes, off_value=eta/classes)
        res.append(lam*y1 + (1-lam)*y2)
    return res

def smooth(label, classes, eta=0.1):
    if isinstance(label, nd.NDArray):
        label = [label]
    smoothed = []
    for l in label:
        res = l.one_hot(classes, on_value=1 - eta + eta/classes, off_value=eta/classes)
        smoothed.append(res)
    return smoothed



import numpy as np


def convert_img(img):
    """Converts (H, W, C) numpy.ndarray to (C, W, H) format
    """
    if len(img.shape) == 3:
        img = img.transpose(2, 0, 1)
    if len(img.shape) == 2:
        img = np.expand_dims(img, 0)
    return img



from collections import OrderedDict
import os
import numpy as np
import torch
from torch import nn

from gluoncv.torch.utils.coot_utils import truncated_normal_fill


class MultiModalTransformer:
    def __init__(self,
                 cfg,
                 use_cuda: bool = True,
                 use_multi_gpu: bool = False):
        self.use_cuda = use_cuda
        self.use_multi_gpu = use_multi_gpu
        self.model_list = []

        self.cfg = cfg
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() and use_cuda else "cpu")
        self.video_pooler = Transformer(
            cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_POOLER,
            cfg.CONFIG.COOT_DATA.FEATURE_DIM)

        self.video_sequencer = Transformer(
            cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_SEQUENCER,
            cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_POOLER.OUTPUT_DIM)

        self.text_pooler = Transformer(
            cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_POOLER,
            cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_ENCODER.FEATURE_DIM)

        self.text_sequencer = Transformer(
            cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_SEQUENCER,
            cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_POOLER.OUTPUT_DIM)

        self.model_list = [
            self.video_pooler, self.video_sequencer, self.text_pooler,
            self.text_sequencer
        ]

    def encode_video(self, vid_frames, vid_frames_mask, vid_frames_len,
                     clip_num, clip_frames, clip_frames_len, clip_frames_mask):
        # compute video context
        vid_context = self.video_pooler(vid_frames, vid_frames_mask,
                                        vid_frames_len, None)
        if self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_SEQUENCER.USE_CONTEXT:
            if self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_SEQUENCER.NAME == "rnn":
                vid_context_hidden = vid_context.unsqueeze(0)
                vid_context_hidden = vid_context_hidden.repeat(
                    self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_SEQUENCER.
                    NUM_LAYERS, 1, 1)
            elif self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_SEQUENCER.NAME == "atn":
                vid_context_hidden = vid_context
            else:
                raise NotImplementedError
        else:
            vid_context_hidden = None

        # compute clip embedding
        clip_emb = self.video_pooler(clip_frames, clip_frames_mask,
                                     clip_frames_len, None)
        batch_size = len(clip_num)
        max_clip_len = torch.max(clip_num)
        clip_feat_dim = self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.VIDEO_POOLER.OUTPUT_DIM
        clip_emb_reshape = torch.zeros(
            (batch_size, max_clip_len, clip_feat_dim))
        clip_emb_mask = torch.zeros((batch_size, max_clip_len))
        clip_emb_lens = torch.zeros((batch_size, ))
        if self.use_cuda:
            clip_emb_reshape = clip_emb_reshape.cuda(non_blocking=True)
            clip_emb_mask = clip_emb_mask.cuda(non_blocking=True)
            clip_emb_lens = clip_emb_lens.cuda(non_blocking=True)
        pointer = 0
        for batch, clip_len in enumerate(clip_num):
            clip_emb_reshape[batch, :clip_len, :] =\
                clip_emb[pointer:pointer + clip_len, :]
            clip_emb_mask[batch, :clip_len] = 1
            clip_emb_lens[batch] = clip_len
            pointer += clip_len

        # compute video embedding
        vid_emb = self.video_sequencer(clip_emb_reshape, clip_emb_mask,
                                       clip_num, vid_context_hidden)

        #TODO: convert the return to an object class or maybe dictionary
        return (vid_emb, clip_emb, vid_context, clip_emb_reshape,
                clip_emb_mask, clip_emb_lens)

    def encode_paragraph(self, paragraph_caption_vectors,
                         paragraph_caption_mask, paragraph_caption_len,
                         sentence_num, sentence_caption_vectors,
                         sentence_caption_mask, sentence_caption_len):
        # compute paragraph context
        paragraph_context = self.text_pooler(paragraph_caption_vectors,
                                             paragraph_caption_mask,
                                             paragraph_caption_len, None)

        if self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_SEQUENCER.USE_CONTEXT:
            if self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_SEQUENCER.NAME == "rnn":
                paragraph_gru_hidden = paragraph_context.unsqueeze(0)
                paragraph_gru_hidden = paragraph_gru_hidden.repeat(
                    self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_SEQUENCER.
                    NUM_LAYERS, 1, 1)
            elif self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_SEQUENCER.NAME == "atn":
                paragraph_gru_hidden = paragraph_context
            else:
                raise NotImplementedError
        else:
            paragraph_gru_hidden = None

        # compute sentence embedding
        sentence_emb = self.text_pooler(sentence_caption_vectors,
                                        sentence_caption_mask,
                                        sentence_caption_len, None)
        batch_size = len(sentence_num)
        sentence_feat_dim = self.cfg.CONFIG.COOT_MODEL.MODEL_CONFIG.TEXT_POOLER.OUTPUT_DIM
        max_sentence_len = torch.max(sentence_num)
        sentence_emb_reshape = torch.zeros(
            (batch_size, max_sentence_len, sentence_feat_dim))
        sentence_emb_mask = torch.zeros((batch_size, max_sentence_len))
        sentence_emb_lens = torch.zeros((batch_size, ))
        if self.use_cuda:
            sentence_emb_reshape = sentence_emb_reshape.cuda(non_blocking=True)
            sentence_emb_mask = sentence_emb_mask.cuda(non_blocking=True)
            sentence_emb_lens = sentence_emb_lens.cuda(non_blocking=True)
        pointer = 0
        for batch, sentence_len in enumerate(sentence_num):
            sentence_emb_reshape[batch, :sentence_len, :] =\
                sentence_emb[pointer:pointer + sentence_len, :]
            sentence_emb_mask[batch, :sentence_len] = 1
            sentence_emb_lens[batch] = sentence_len
            pointer += sentence_len

        # compute paragraph embedding
        paragraph_emb = self.text_sequencer(sentence_emb_reshape,
                                            sentence_emb_mask, sentence_num,
                                            paragraph_gru_hidden)

        return (paragraph_emb, sentence_emb, paragraph_context,
                sentence_emb_reshape, sentence_emb_mask, sentence_emb_lens)

    def eval(self):
        for model in self.model_list:
            model.eval()
        torch.set_grad_enabled(False)

    def cuda(self, gpu):
        for model in self.model_list:
            self._to_device_fn(model, gpu)

    def train(self):
        for model in self.model_list:
            model.train()
        torch.set_grad_enabled(True)

    def _to_device_fn(self, model, gpu=0):
        torch.cuda.set_device(gpu)
        if self.use_multi_gpu:
            model = nn.DataParallel(model)
        model = model.to(self.device)
        return model

    def get_params(self):
        params = []
        for model in self.model_list:
            params_dict = OrderedDict(model.named_parameters())
            _params = []
            for key, value in params_dict.items():
                _params += [{'params': value}]
            params.extend(_params)
        return params

    def load_checkpoint(self, ckpt: str):
        state = torch.load(str(ckpt))
        for i, model in enumerate(self.model_list):
            state_dict = state[i]
            if self.use_multi_gpu:
                newer_state_dict = OrderedDict()
                for key, val in state_dict.items():
                    assert not key.startswith("module.")
                    new_key = "module." + key
                    newer_state_dict[new_key] = val
                model.load_state_dict(newer_state_dict)
            else:
                model.load_state_dict(state_dict)
            i += 1  # we do this intentionally

    def save_model(self, optimizer, epoch, cfg):
        """
        Save trained model weights.
        """
        model_save_dir = os.path.join(cfg.CONFIG.LOG.BASE_PATH,
                                      cfg.CONFIG.LOG.EXP_NAME,
                                      cfg.CONFIG.LOG.SAVE_DIR)
        if not os.path.exists(model_save_dir):
            os.makedirs(model_save_dir)
        ckpt_name = "f{}_s{}_ckpt_epoch{}.pth".format(
            cfg.CONFIG.DATA.CLIP_LEN, cfg.CONFIG.DATA.FRAME_RATE, epoch)
        checkpoint = os.path.join(model_save_dir, ckpt_name)
        model_states = []
        for m in self.model_list:
            state_dict = m.state_dict()
            if self.use_multi_gpu:
                new_state_dict = OrderedDict()
                for key, val in state_dict.items():
                    assert key.startswith("module.")
                    new_key = key[7:]
                    new_state_dict[new_key] = val
                model_states.append(new_state_dict)
            else:
                model_states.append(state_dict)
        state = {
            'epoch': epoch + 1,
            'state_dict': model_states,
            'best_acc1': None,
            'optimizer': optimizer.state_dict(),
        }
        torch.save(state, checkpoint)


class LayerNormalization(nn.Module):
    def __init__(self, features_count, epsilon=1e-6):
        super().__init__()
        self.gain = nn.Parameter(torch.ones(features_count),
                                 requires_grad=True)
        self.bias = nn.Parameter(torch.zeros(features_count),
                                 requires_grad=True)
        self.epsilon = epsilon

    def forward(self, x):
        mean = x.mean(dim=-1, keepdim=True)
        std = x.std(dim=-1, keepdim=True)
        return self.gain * (x - mean) / (std + self.epsilon) + self.bias


def parse_pooler(input_dim, cfg) -> nn.Module:
    if cfg.POOLER == "atn":
        return AtnPool(input_dim, cfg.ATN_POOL_DIM, cfg.ATN_POOL_HEADS,
                       cfg.DROPOUT)
    elif cfg.POOLER == "avg":
        return AvgPool()
    elif cfg.POOLER == "max":
        return MaxPool()
    else:
        raise ValueError(f"unknown pooler {cfg.POOLER}")


class Transformer(nn.Module):
    def __init__(self, cfg, feature_dim: int):
        super(Transformer, self).__init__()

        self.input_norm = LayerNormalization(feature_dim)
        self.input_fc = None
        input_dim = feature_dim

        if cfg.INPUT_FC:
            self.input_fc = nn.Sequential(
                nn.Linear(feature_dim, cfg.INPUT_FC_OUTPUT_DIM), nn.GELU())
            input_dim = cfg.INPUT_FC_OUTPUT_DIM
        self.embedding = PositionalEncoding(input_dim,
                                            cfg.DROPOUT,
                                            max_len=1000)

        self.tf = TransformerEncoder(cfg.NUM_LAYERS, input_dim, cfg.NUM_HEADS,
                                     input_dim, cfg.DROPOUT)

        self.use_context = cfg.USE_CONTEXT
        if self.use_context:
            self.tf_context = TransformerEncoder(cfg.ATN_CTX_NUM_LAYERS,
                                                 input_dim,
                                                 cfg.ATN_CTX_NUM_HEADS,
                                                 input_dim, cfg.DROPOUT)

        self.pooler = parse_pooler(input_dim, cfg)

        init_network(self, init_std=0.01)

    def forward(self, features, mask, lengths, hidden_state):
        features = self.input_norm(features)
        if self.input_fc is not None:
            features = self.input_fc(features)
        features = self.embedding(features)
        features = self.tf(features, features, features, mask)
        add_after_pool = None
        if self.use_context:
            hidden_state = hidden_state.unsqueeze(1)
            ctx = self.tf_context(hidden_state, features, features, mask)
            add_after_pool = ctx.squeeze(1)
        pooled = self.pooler(features, mask, lengths)
        if add_after_pool is not None:
            pooled = torch.cat([pooled, add_after_pool], dim=-1)
        return pooled


class PositionalEncoding(nn.Module):
    def __init__(self, dim, dropout_prob=0., max_len=1000):
        super(PositionalEncoding, self).__init__()
        pe = torch.zeros(max_len, dim).float()
        position = torch.arange(0, max_len).unsqueeze(1).float()
        dimension = torch.arange(0, dim).float()
        div_term = 10000**(2 * dimension / dim)
        pe[:, 0::2] = torch.sin(position / div_term[0::2])
        pe[:, 1::2] = torch.cos(position / div_term[1::2])
        self.register_buffer('pe', pe)
        self.dropout = nn.Dropout(p=dropout_prob)
        self.dim = dim

    def forward(self, x, step=None):
        if step is None:
            x = x + self.pe[:x.size(1), :]
        else:
            x = x + self.pe[:, step]
        return self.dropout(x)


class TransformerEncoder(nn.Module):
    def __init__(self, layers_count, model_dim, heads_count, fc_dim,
                 dropout_prob):
        super(TransformerEncoder, self).__init__()
        self.model_dim = model_dim
        assert layers_count > 0
        self.encoder_layers = nn.ModuleList([
            TransformerEncoderLayer(model_dim, heads_count, fc_dim,
                                    dropout_prob) for _ in range(layers_count)
        ])

    def forward(self, query, key, value, mask):
        batch_size, query_len, embed_dim = query.shape
        batch_size, key_len, embed_dim = key.shape
        mask = (1 - mask.unsqueeze(1).expand(batch_size, query_len, key_len))
        mask = mask == 1
        sources = None

        sources = self.encoder_layers[0](query, key, value, mask)
        return sources


class TransformerEncoderLayer(nn.Module):
    def __init__(self, model_dim, heads_count, fc_dim, dropout_prob):
        super(TransformerEncoderLayer, self).__init__()
        self.self_attention_layer = Sublayer(
            MultiHeadAttention(heads_count, model_dim, dropout_prob),
            model_dim)
        self.pointwise_feedforward_layer = Sublayer(
            PointwiseFeedForwardNetwork(fc_dim, model_dim, dropout_prob),
            model_dim)
        self.dropout = nn.Dropout(dropout_prob)

    def forward(self, query, key, value, sources_mask):
        sources = self.self_attention_layer(query, key, value, sources_mask)
        sources = self.dropout(sources)
        sources = self.pointwise_feedforward_layer(sources)
        return sources


class Sublayer(nn.Module):
    def __init__(self, sublayer, model_dim):
        super(Sublayer, self).__init__()
        self.sublayer = sublayer
        self.layer_normalization = LayerNormalization(model_dim)

    def forward(self, *args):
        x = args[0]
        x = self.sublayer(*args) + x
        return self.layer_normalization(x)


class MultiHeadAttention(nn.Module):
    def __init__(self, heads_count, model_dim, dropout_prob):
        super(MultiHeadAttention, self).__init__()
        assert model_dim % heads_count == 0,\
            f"model dim {model_dim} not divisible by {heads_count} heads"
        self.d_head = model_dim // heads_count
        self.heads_count = heads_count
        self.query_projection = nn.Linear(model_dim, heads_count * self.d_head)
        self.key_projection = nn.Linear(model_dim, heads_count * self.d_head)
        self.value_projection = nn.Linear(model_dim, heads_count * self.d_head)
        self.final_projection = nn.Linear(model_dim, heads_count * self.d_head)
        self.dropout = nn.Dropout(dropout_prob)
        self.softmax = nn.Softmax(dim=3)
        self.attention = None

    def forward(self, query, key, value, mask=None):
        batch_size, query_len, model_dim = query.size()
        d_head = model_dim // self.heads_count
        query_projected = self.query_projection(query)
        key_projected = self.key_projection(key)
        value_projected = self.value_projection(value)
        batch_size, key_len, model_dim = key_projected.size()
        batch_size, value_len, model_dim = value_projected.size()
        query_heads = query_projected.view(batch_size, query_len,
                                           self.heads_count,
                                           d_head).transpose(1, 2)
        key_heads = key_projected.view(batch_size, key_len, self.heads_count,
                                       d_head).transpose(1, 2)
        value_heads = value_projected.view(batch_size, value_len,
                                           self.heads_count,
                                           d_head).transpose(1, 2)
        attention_weights = self.scaled_dot_product(query_heads, key_heads)
        if mask is not None:
            mask_expanded = mask.unsqueeze(1).expand_as(attention_weights)
            attention_weights = attention_weights.masked_fill(
                mask_expanded, -1e18)
        attention = self.softmax(attention_weights)
        attention_dropped = self.dropout(attention)
        context_heads = torch.matmul(attention_dropped, value_heads)
        context_sequence = context_heads.transpose(1, 2)
        context = context_sequence.reshape(batch_size, query_len, model_dim)
        final_output = self.final_projection(context)
        return final_output

    def scaled_dot_product(self, query_heads, key_heads):
        key_heads_transposed = key_heads.transpose(2, 3)
        dot_product = torch.matmul(query_heads, key_heads_transposed)
        attention_weights = dot_product / np.sqrt(self.d_head)
        return attention_weights


class PointwiseFeedForwardNetwork(nn.Module):
    def __init__(self, fc_dim, model_dim, dropout_prob):
        super(PointwiseFeedForwardNetwork, self).__init__()
        self.feed_forward = nn.Sequential(nn.Linear(model_dim, fc_dim),
                                          nn.Dropout(dropout_prob), nn.GELU(),
                                          nn.Linear(fc_dim, model_dim),
                                          nn.Dropout(dropout_prob))

    def forward(self, x):
        return self.feed_forward(x)


class AvgPool(nn.Module):
    def forward(self, features, mask, lengths):
        _ = mask
        len_div = lengths.unsqueeze(-1).float()
        result_sum = torch.sum(features, dim=1)
        result = result_sum / len_div
        return result


class MaxPool(nn.Module):
    def forward(self, features, mask, lengths):
        result_max, _ = torch.max(features, dim=1)
        return result_max


class AtnPool(nn.Module):
    def __init__(self, d_input, d_attn, n_heads, dropout_prob):
        super(AtnPool, self).__init__()
        self.d_head = d_attn // n_heads
        self.d_head_output = d_input // n_heads
        self.num_heads = n_heads

        def _init(tensor_):
            tensor_.data = (truncated_normal_fill(tensor_.data.shape,
                                                  std=0.01))

        w1_head = torch.zeros(n_heads, d_input, self.d_head)
        b1_head = torch.zeros(n_heads, self.d_head)
        w2_head = torch.zeros(n_heads, self.d_head, self.d_head_output)
        b2_head = torch.zeros(n_heads, self.d_head_output)
        _init(w1_head)
        _init(b1_head)
        _init(w2_head)
        _init(b2_head)
        self.genpool_w1_head = nn.Parameter(w1_head, requires_grad=True)
        self.genpool_b1_head = nn.Parameter(b1_head, requires_grad=True)
        self.genpool_w2_head = nn.Parameter(w2_head, requires_grad=True)
        self.genpool_b2_head = nn.Parameter(b2_head, requires_grad=True)
        self.activation = nn.GELU()
        self.dropout1 = nn.Dropout(dropout_prob)
        self.dropout2 = nn.Dropout(dropout_prob)
        self.dropout3 = nn.Dropout(dropout_prob)
        self.softmax = nn.Softmax(dim=2)
        self.softmax_temp = 1
        self.genpool_one = nn.Parameter(torch.ones(1), requires_grad=False)

    def extra_repr(self) -> str:
        strs = []
        for p in [
                self.genpool_w1_head, self.genpool_b1_head,
                self.genpool_w2_head, self.genpool_b2_head
        ]:
            strs.append(f"pool linear {p.shape}")
        return "\n".join(strs)

    def forward(self, features, mask, lengths):

        batch_size, seq_len, input_dim = features.shape
        b1 = torch.matmul(features.unsqueeze(1),
                          self.genpool_w1_head.unsqueeze(0))
        b1 += self.genpool_b1_head.unsqueeze(1).unsqueeze(0)
        b1 = self.activation(self.dropout1(b1))
        b1 = torch.matmul(b1, self.genpool_w2_head.unsqueeze(0))
        b1 += self.genpool_b2_head.unsqueeze(1).unsqueeze(0)
        b1 = self.dropout2(b1)
        b1.masked_fill_((mask == 0).unsqueeze(1).unsqueeze(-1), -1e19)

        smweights = self.softmax(b1 / self.softmax_temp)
        smweights = self.dropout3(smweights)
        smweights = smweights.transpose(1, 2).reshape(-1, seq_len, input_dim)
        return (features * smweights).sum(
            dim=1)  # pool features with attention weights


def _init_weight(w, init_gain=1):
    w.copy_(truncated_normal_fill(w.shape, std=init_gain))


def init_network(net: nn.Module, init_std: float):
    for key, val in net.named_parameters():
        if "weight" in key or "bias" in key:
            _init_weight(val.data, init_std)



import os
import random
import logging
import ray
from pathlib import Path
from tqdm import tqdm

from .io.video_io import VideoFrameReader, write_img_files_to_vid_ffmpeg, \
    resize_and_write_video_ffmpeg, convert_vid_ffmpeg
from .utils.serialization_utils import save_json
from .dataset import GluonCVMotionDataset, DataSample, FieldNames, get_resized_video_location, \
    get_vis_gt_location, get_vis_thumb_location, get_vis_video_location


def generate_resized_video(data_sample:DataSample, short_edge_res:int, overwrite=False, cache_dir=None,
                           upscale=True, use_vis_path=True, force_encode=False, encode_kwargs=None, **kwargs):
    if cache_dir is None:
        resized_path = get_resized_video_location(data_sample,short_edge_res)
    else:
        resized_path = data_sample.get_cache_file(cache_dir, extension='.mp4')
    if os.path.isfile(resized_path) and not overwrite:
        return
    os.makedirs(os.path.dirname(resized_path), exist_ok=True)

    orig_vid_path = get_vis_video_location(data_sample) if use_vis_path else data_sample.data_path
    if not upscale and min(data_sample.width, data_sample.height) <= short_edge_res:
        if force_encode:
            if encode_kwargs is None:
                encode_kwargs = {k: v for k, v in kwargs.items() if k == "addn_args"}
            convert_vid_ffmpeg(orig_vid_path, resized_path, **encode_kwargs)
        else:
            os.symlink(orig_vid_path, resized_path)
    else:
        resize_and_write_video_ffmpeg(orig_vid_path, resized_path, short_edge_res, **kwargs)

    return resized_path


def generate_video(data_sample:DataSample, force_encode=False, overwrite=False, **kwargs):
    new_file = get_vis_video_location(data_sample)
    if (os.path.isfile(new_file) and not overwrite) and not (force_encode and os.path.islink(new_file)):
        return
    os.makedirs(os.path.dirname(new_file), exist_ok=True)

    video_file = data_sample.data_path
    #### Generate Video ####
    if os.path.isdir(video_file):
        # the data is a set of images, generate a video
        img_files = [os.path.join(video_file, f) for f in sorted(os.listdir(video_file))]
        write_img_files_to_vid_ffmpeg(out_file=new_file, in_files=img_files, fps=data_sample.metadata['fps'])
    else:
        if not video_file.endswith(".mp4") or force_encode:
            # Convert the video to mp4
            convert_vid_ffmpeg(video_file, new_file, **kwargs)
        else:
            # the data is a video, symlink to it
            if os.path.exists(new_file):
                os.remove(new_file)
            os.symlink(video_file, new_file)

    return new_file


def generate_thumbnail(data_sample:DataSample, overwrite=False):
    video_thumbnail_frame = get_vis_thumb_location(data_sample)
    if os.path.isfile(video_thumbnail_frame) and not overwrite:
        return
    os.makedirs(os.path.dirname(video_thumbnail_frame), exist_ok=True)

    video_file = get_vis_video_location(data_sample)

    #### Generate Thumbnail ####
    vid = VideoFrameReader(video_file)
    img, ts = vid.get_frame(30) if len(vid) > 30 else vid.get_frame(0)
    img.thumbnail((300, 300))
    img.save(video_thumbnail_frame)

    return video_thumbnail_frame


def generate_gt_vis_json(data_sample:DataSample, cache_suffix="", overwrite=False):
    gt_file = get_vis_gt_location(data_sample, cache_suffix)
    if os.path.isfile(gt_file) and not overwrite:
        return
    #### Generate GT Track json ####
    os.makedirs(os.path.dirname(gt_file), exist_ok=True)

    vis_video_file = get_vis_video_location(data_sample)
    vis_vid = VideoFrameReader(vis_video_file)
    sample_dict = data_sample.to_dict(include_id=True)
    sample_dict[FieldNames.METADATA][FieldNames.FPS] = vis_vid.fps
    save_json(sample_dict, gt_file, indent=0)

    return gt_file


@ray.remote
def generate_files_for_one_sample_ray(data_sample:DataSample, generator_list, overwrite):
    generate_files_for_one_sample(data_sample, generator_list, overwrite)

@ray.remote
def generate_files_for_multi_samples_ray(data_samples, generator_list, overwrite):
    for id, data_sample in data_samples:
        generate_files_for_one_sample(data_sample, generator_list, overwrite)
        logging.info('Finished: {}'.format(data_sample.data_path))

def generate_files_for_one_sample(data_sample:DataSample, generator_list, overwrite):
    if os.path.isabs(data_sample.data_relative_path):
        logging.error("Relative path of sample id: {} is absolute and so we cannot add to the cache, skipping."
                      " Path: {}".format(data_sample.id, data_sample.data_relative_path))
        return
    for gen in generator_list:
        try:
            gen(data_sample, overwrite)
        except Exception as e:
            try:
                gen(data_sample, True)
            except Exception as e:
                logging.exception('Failed: {}'.format(data_sample.data_path))
                return


def generate_preprocess_files(part=0, parts=1,
                              annotation_file='./kinetics/annotation/anno_400.json',
                              use_ray=False, num_cpus=4, overwrite=False, force_encode=False, short_edge_res=256,
                              distributed=False, shuffle_seed=None, dataset=None):

    if dataset is None:
        dataset = GluonCVMotionDataset(annotation_file)

    generator_list = [
        lambda sample, overwrite: generate_video(sample, force_encode, overwrite),
        generate_thumbnail,
        lambda sample, overwrite: generate_gt_vis_json(sample, dataset.get_anno_subpath(), overwrite),
    ]

    samples = sorted(dataset.samples)
    if shuffle_seed is not None and shuffle_seed != "None":
        random.seed(shuffle_seed)
        random.shuffle(samples)
    samples = samples[part::parts]

    logging.info("Using ray {} ,  distributed: {}".format(use_ray, distributed))
    if use_ray:
        num_ray_threads = 500
        if distributed:
            ray.init(redis_address="localhost:6379")
        else:
            ray.init(num_cpus=num_cpus)
        ray.get([generate_files_for_multi_samples_ray.remote(samples[i::num_ray_threads], generator_list, overwrite) for i in range(num_ray_threads)])
    else:
        for id, data_sample in tqdm(samples, mininterval=1.0):
            generate_files_for_one_sample(data_sample, generator_list, overwrite)


if __name__ == '__main__':
    import fire
    fire.Fire(generate_preprocess_files)



"""IOU loss"""
import torch
from torch import nn


class IOULoss(nn.Module):
    """
    Intersetion Over Union (IoU) loss which supports three
    different IoU computations:

    * IoU
    * Linear IoU
    * gIoU
    """
    def __init__(self, loc_loss_type='iou'):
        super(IOULoss, self).__init__()
        self.loc_loss_type = loc_loss_type

    def forward(self, pred, target, weight=None):
        """
        Args:
            pred: Nx4 predicted bounding boxes
            target: Nx4 target bounding boxes
            weight: N loss weight for each instance
        """
        pred_left = pred[:, 0]
        pred_top = pred[:, 1]
        pred_right = pred[:, 2]
        pred_bottom = pred[:, 3]

        target_left = target[:, 0]
        target_top = target[:, 1]
        target_right = target[:, 2]
        target_bottom = target[:, 3]

        target_aera = (target_left + target_right) * \
                      (target_top + target_bottom)
        pred_aera = (pred_left + pred_right) * \
                    (pred_top + pred_bottom)

        w_intersect = torch.min(pred_left, target_left) + \
                      torch.min(pred_right, target_right)
        h_intersect = torch.min(pred_bottom, target_bottom) + \
                      torch.min(pred_top, target_top)

        g_w_intersect = torch.max(pred_left, target_left) + \
                        torch.max(pred_right, target_right)
        g_h_intersect = torch.max(pred_bottom, target_bottom) + \
                        torch.max(pred_top, target_top)
        ac_uion = g_w_intersect * g_h_intersect

        area_intersect = w_intersect * h_intersect
        area_union = target_aera + pred_aera - area_intersect

        ious = (area_intersect + 1.0) / (area_union + 1.0)
        gious = ious - (ac_uion - area_union) / ac_uion
        if self.loc_loss_type == 'iou':
            losses = -torch.log(ious)
        elif self.loc_loss_type == 'linear_iou':
            losses = 1 - ious
        elif self.loc_loss_type == 'giou':
            losses = 1 - gious
        else:
            raise NotImplementedError

        if weight is not None:
            return (losses * weight).sum()
        else:
            return losses.sum()



"""Base Estimator"""
# pylint: disable=bare-except
import os
import math
import pickle
import io
import logging
from datetime import datetime
import numpy as np
import pandas as pd
from ...utils import random as _random
from ...utils.filesystem import temporary_filename
from .utils import _suggest_load_context

logging.basicConfig(level=logging.INFO)

def set_default(cfg):
    """A special hook to register the default values for decorated Estimator.

    Parameters
    ----------
    cfg : autocfg.dataclass
        Default config as dataclass object
    """
    def _apply(cls):
        # docstring
        cls.__doc__ = str(cls.__doc__) if cls.__doc__ else ''
        cls.__doc__ += ("\n\nParameters\n"
                        "----------\n"
                        "config : str, dict\n"
                        "  Config used to override default configurations. \n"
                        "  If `str`, assume config file (.yml, .yaml) is used. \n"
                        "logger : logger, default is `None`.\n"
                        "  If not `None`, will use default logging object.\n")
        cls.__doc__ += '\nDefault configurations: \n--------------------\n'
        sio = io.StringIO()
        cfg.save(sio)
        cls.__doc__ += '\n' + sio.getvalue()
        cls._default_cfg = cfg
        cls._default_cfg.freeze()
        return cls
    return _apply


class BaseEstimator:
    """This is the base estimator for gluoncv.auto.Estimators.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable
        The reporter for metric checkpointing.
    name : str
        Optional name for the estimator.

    Attributes
    ----------
    _logger : logging.Logger
        The customized/default logger for this estimator.
    _logdir : str
        The temporary dir for logs.
    _cfg : autocfg.dataclass
        The configurations.

    """
    def __init__(self, config, logger=None, reporter=None, name=None):
        self._reporter = reporter
        name = name if isinstance(name, str) else self.__class__.__name__
        self._name = name
        self._logger = logger if logger is not None else logging.getLogger(name)
        self._logger.setLevel(logging.INFO)

        # reserved attributes
        self.net = None
        self.num_class = None
        self.classes = []
        self.ctx = [None]
        self.dataset = 'auto'

        # logdir
        logdir = config.pop('log_dir', None) if isinstance(config, dict) else None
        if logdir:
            self._logdir = os.path.abspath(logdir)
        else:
            self._logdir = os.path.join(os.getcwd(), name.lower() + datetime.now().strftime("-%m-%d-%Y"))

        # finalize config
        cfg = self._default_cfg.merge(config)  # config can be dict or yaml file
        diffs = self._default_cfg.diff(cfg)
        if diffs:
            self._logger.info('modified configs(<old> != <new>): {')
            for diff in diffs:
                self._logger.info(diff)
            self._logger.info('}')
        self._cfg = cfg

        os.makedirs(self._logdir, exist_ok=True)
        config_file = os.path.join(self._logdir, 'config.yaml')
        # log file
        self._log_file = os.path.join(self._logdir, 'estimator.log')
        fh = logging.FileHandler(self._log_file)
        self._logger.addHandler(fh)
        # save_config(r.config, self._logger, config_file)
        self._cfg.save(config_file)
        self._logger.info('Saved config to %s', config_file)

        # freeze config
        self._cfg.freeze()
        seed = self._cfg.get('seed', np.random.randint(1000000))
        _random.seed(seed)

    def fit(self, train_data, val_data=None, train_size=0.9, random_state=None,
            resume=False, time_limit=None):

        """Fit with train/validation data.

        Parameters
        ----------
        train_data : pd.DataFrame or iterator
            Training data.
        val_data : pd.DataFrame or iterator, optional
            Validation data, optional. If `train_data` is DataFrame, `val_data` will be split from
            `train_data` given `train_size`.
        train_size : float
            The portion of train data split from original `train_data` if `val_data` is not provided.
        random_state : int
            Random state for splitting, for `np.random.seed`.
        resume : bool
            Whether resume from previous `fit`(if possible) or start as fresh.
        time_limit : int, default is None
            The wall clock time limit(second) for fit process, if `None`, time limit is not enforced.
            If `fit` takes longer than `time_limit`, the process will terminate early and return the
            model prematurally.
            Due to callbacks and additional validation functions, the `time_limit` may not be very precise
            (few minutes allowance), but you can use it to safe-guard a very long training session.

        Returns
        -------
        None

        """
        if time_limit is None:
            time_limit = math.inf
        elif not isinstance(time_limit, (int, float)):
            raise TypeError(f'Invalid type `time_limit={time_limit}`, int/float or None expected')
        if not resume:
            self.classes = train_data.classes
            self.num_class = len(self.classes)
            self._init_network()
        if not isinstance(train_data, pd.DataFrame):
            assert val_data is not None, \
                "Please provide `val_data` as we do not know how to split `train_data` of type: \
                {}".format(type(train_data))
            ret = self._fit(train_data, val_data, time_limit=time_limit) if not resume \
                else self._resume_fit(train_data, val_data, time_limit=time_limit)
            return self._reload_best(ret)

        os.makedirs(self._logdir, exist_ok=True)
        if val_data is None:
            assert 0 <= train_size <= 1.0
            if random_state:
                np.random.seed(random_state)
            split_mask = np.random.rand(len(train_data)) < train_size
            train = train_data[split_mask]
            val = train_data[~split_mask]
            self._logger.info('Randomly split train_data into train[%d]/validation[%d] splits.',
                              len(train), len(val))
            ret = self._fit(train, val, time_limit=time_limit) if not resume else \
                self._resume_fit(train, val, time_limit=time_limit)
            return self._reload_best(ret)

        ret = self._fit(train_data, val_data, time_limit=time_limit) if not resume else \
            self._resume_fit(train_data, val_data, time_limit=time_limit)
        return self._reload_best(ret)

    def evaluate(self, val_data):
        """Evaluate estimator on validation data.

        Parameters
        ----------
        val_data : pd.DataFrame or iterator
            The validation data.

        """
        return self._evaluate(val_data)

    def predict(self, x, **kwargs):
        """Predict using this estimator.

        Parameters
        ----------
        x : str, pd.DataFrame or ndarray
            The input, can be str(filepath), pd.DataFrame with 'image' column, or raw ndarray input.

        """
        return self._predict(x, **kwargs)

    def predict_feature(self, x):
        """Predict intermediate features using this estimator.

        Parameters
        ----------
        x : str, pd.DataFrame or ndarray
            The input, can be str(filepath), pd.DataFrame with 'image' column, or raw ndarray input.

        """
        return self._predict_feature(x)

    def _reload_best(self, return_value):
        """Applying the best checkpoint before return"""
        cp = return_value.get('checkpoint', '')
        if not cp:
            return return_value
        self._logger.info('Applying the state from the best checkpoint...')
        try:
            tmp = self.load(cp)
            self.__dict__.update(tmp.__dict__)
        except:
            self._logger.warning(
                'Unable to resume the state from the best checkpoint, using the latest state.')
        return return_value

    def _predict(self, x, **kwargs):
        raise NotImplementedError

    def _predict_feature(self, x, **kwargs):
        raise NotImplementedError

    def _fit(self, train_data, val_data, time_limit=math.inf):
        raise NotImplementedError

    def _resume_fit(self, train_data, val_data, time_limit=math.inf):
        raise NotImplementedError

    def _evaluate(self, val_data):
        raise NotImplementedError

    def _init_network(self, **kwargs):
        raise NotImplementedError

    def _init_trainer(self):
        raise NotImplementedError

    def _validate_gpus(self, gpu_ids):
        """validate if requested gpus are actually available"""
        valid_gpus = []
        try:
            import mxnet as mx
            for gid in gpu_ids:
                try:
                    _ = mx.nd.zeros((1,), ctx=mx.gpu(gid))
                    valid_gpus.append(gid)
                except:
                    pass
        except ImportError:
            pass
        return valid_gpus

    #FIXME: better design than a duplicate function?
    def _torch_validate_gpus(self, gpu_ids):
        """validate if requested gpus are actually available"""
        valid_gpus = []
        try:
            import torch
            if not torch.cuda.is_available():
                return valid_gpus
            for gid in gpu_ids:
                try:
                    _ = torch.zeros(1, device=f'cuda:{gid}')
                    valid_gpus.append(str(gid))
                except:
                    pass
        except ImportError:
            pass
        return valid_gpus

    def reset_ctx(self, ctx=None):
        """Reset model context.

        Parameters
        ----------
        ctx : list or ctx
            The desired new ctx list, the type must match the network.
            For example, if using mxnet, the ctx must be `mxnet.Context`.

        """
        if not ctx:
            return
        if not isinstance(ctx, (tuple, list)):
            ctx_list = [ctx]
        else:
            ctx_list = ctx
        done = False
        try:
            import mxnet as mx
            if isinstance(self.net, mx.gluon.Block):
                for c in ctx_list:
                    assert isinstance(c, mx.Context)
                if hasattr(self.net, 'reset_ctx'):
                    self.net.reset_ctx(ctx_list)
                else:
                    self.net.collect_params().reset_ctx(ctx_list)
                self.ctx = ctx_list
                done = True
        except ImportError:
            pass
        try:
            import torch
            if isinstance(self.net, (torch.nn.Module, torch.nn.DataParallel)):
                for c in ctx_list:
                    assert isinstance(c, torch.device)
                if hasattr(self.net, 'reset_ctx'):
                    self.net.reset_ctx(ctx_list)
                else:
                    if isinstance(self.net, torch.nn.DataParallel):
                        self.net = torch.nn.DataParallel(self.net.module, device_ids=[ctx.index for ctx in ctx_list])
                    self.net.to(self.ctx[0])
                self.ctx = ctx_list
                done = True
        except ImportError:
            pass
        if not done:
            raise RuntimeError("Unable to reset_ctx, no `mxnet` and `pytorch`.")

    def save(self, filename):
        """Save the state of this estimator to disk.

        Parameters
        ----------
        filename : str
            The file name for storing the full state.
        """
        with open(filename, 'wb') as fid:
            try:
                pickle.dump(self, fid)
            except pickle.PicklingError as e:
                self._logger.warning(f"Unable to pickle object due to the reason: {str(e)}. This object is not saved.")
                return
        self._logger.debug('Pickled to %s', filename)

    @classmethod
    def load(cls, filename, ctx='auto'):
        """Load the state from disk copy.

        Parameters
        ----------
        filename : str
            The file name to load from.
        ctx: str, default is 'auto'
            The context for reloaded model.
            'auto': use previously saved context type if still available, fallback
            to cpu if no gpu detected.
            Use `cpu` if no GPU available.
            'cpu': use cpu for inference regardless.
            'gpu': use as many gpus available as possible.
            [0, 2, 4, ...]: if a list or tuple of integers are provided, the context
            will be [gpu(0), gpu(2), gpu(4)...]
        """
        with open(filename, 'rb') as fid:
            obj = pickle.load(fid)
            obj._logger.debug('Unpickled from %s', filename)
            new_ctx = _suggest_load_context(obj.net, ctx, obj.ctx)
            obj.reset_ctx(new_ctx)
            return obj

    def __getstate__(self):
        d = self.__dict__.copy()
        try:
            import mxnet as mx
            d.pop('async_net', None)
            d.pop('_feature_net', None)
            net = d.get('net', None)
            if isinstance(net, mx.gluon.HybridBlock):
                with temporary_filename() as tfile:
                    net.save_parameters(tfile)
                    with open(tfile, 'rb') as fi:
                        d['net'] = fi.read()
            trainer = d.get('trainer', None)
            if isinstance(trainer, mx.gluon.Trainer):
                with temporary_filename() as tfile:
                    trainer.save_states(tfile)
                    with open(tfile, 'rb') as fi:
                        d['trainer'] = fi.read()
        except ImportError:
            pass
        d['_logger'] = None
        d['_reporter'] = None
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        # logger
        self._logger = logging.getLogger(state.get('_name', self.__class__.__name__))
        self._logger.setLevel(logging.ERROR)
        try:
            fh = logging.FileHandler(self._log_file)
            self._logger.addHandler(fh)
        #pylint: disable=bare-except
        except:
            pass
        try:
            import mxnet as _
            net_params = state['net']
            self._init_network(load_only=True)
            with temporary_filename() as tfile:
                with open(tfile, 'wb') as fo:
                    fo.write(net_params)
                self.net.load_parameters(tfile, ignore_extra=True)
            trainer_state = state['trainer']
            self._init_trainer()
            with temporary_filename() as tfile:
                with open(tfile, 'wb') as fo:
                    fo.write(trainer_state)
                self.trainer.load_states(tfile)
        except ImportError:
            pass
        self._logger.setLevel(logging.INFO)



"""COCO evaluation utils"""
# pylint: disable=line-too-long, redefined-builtin
import contextlib
import copy
import io
import itertools
import json
import logging
import os
import pickle
from collections import OrderedDict

import numpy as np
import torch
import pandas as pd

from .. import comm
from ...data.registry.catalog import MetadataCatalog
from ...data.registry.mscoco import convert_to_coco_json
from ...data.structures import Boxes, BoxMode, pairwise_iou
from ...data.detection.detection_utils import try_import_pycocotools


class COCOEvaluator:
    """
    Evaluate AR for object proposals, AP for instance detection/segmentation, AP
    for keypoint detection outputs using COCO's metrics.
    See http://cocodataset.org/#detection-eval and
    http://cocodataset.org/#keypoints-eval to understand its metrics.

    In addition to COCO, this evaluator is able to support any bounding box detection,
    instance segmentation, or keypoint detection dataset.
    """

    def __init__(self, dataset_name, cfg, distributed, output_dir=None):
        """
        Args:
            dataset_name (str): name of the dataset to be evaluated.
                It must have either the following corresponding metadata:

                    "json_file": the path to the COCO format annotation

                Or it must be in detectron2's standard dataset format
                so it can be converted to COCO format automatically.
            cfg (CfgNode): config instance
            distributed (True): if True, will collect results from all ranks and run evaluation
                in the main process.
                Otherwise, will evaluate the results in the current process.
            output_dir (str): optional, an output directory to dump all
                results predicted on the dataset. The dump contains two files:

                1. "instance_predictions.pth" a file in torch serialization
                   format that contains all the raw original predictions.
                2. "coco_instances_results.json" a json file in COCO's result
                   format.
        """
        try_import_pycocotools()
        from pycocotools.coco import COCO
        self._tasks = self._tasks_from_config(cfg)
        self._distributed = distributed
        self._output_dir = output_dir

        self._cpu_device = torch.device("cpu")
        self._logger = logging.getLogger(__name__)

        self._metadata = MetadataCatalog.get(dataset_name)
        if not hasattr(self._metadata, "json_file"):
            self._logger.info(
                f"'{dataset_name}' is not registered by `register_coco_instances`."
                " Therefore trying to convert it to COCO format ..."
            )

            cache_path = os.path.join(output_dir, f"{dataset_name}_coco_format.json")
            self._metadata.json_file = cache_path
            convert_to_coco_json(dataset_name, cache_path)

        json_file = self._metadata.json_file
        with contextlib.redirect_stdout(io.StringIO()):
            self._coco_api = COCO(json_file)

        self._kpt_oks_sigmas = cfg.CONFIG.VAL.KEYPOINT_OKS_SIGMAS
        # Test set json files do not contain annotations (evaluation must be
        # performed using the COCO evaluation server).
        self._do_evaluation = "annotations" in self._coco_api.dataset

    def reset(self):
        self._predictions = []

    def _tasks_from_config(self, cfg):
        """
        Returns:
            tuple[str]: tasks that can be evaluated under the given configuration.
        """
        tasks = ("bbox",)
        if cfg.CONFIG.DATA.MASK_ON:
            tasks = tasks + ("segm",)
        if cfg.CONFIG.DATA.KEYPOINT_ON:
            tasks = tasks + ("keypoints",)
        return tasks

    def process(self, inputs, outputs):
        """
        Args:
            inputs: the inputs to a COCO model (e.g., GeneralizedRCNN).
                It is a list of dict. Each dict corresponds to an image and
                contains keys like "height", "width", "file_name", "image_id".
            outputs: the outputs of a COCO model. It is a list of dicts with key
                "instances" that contains :class:`Instances`.
        """
        for input, output in zip(inputs, outputs):
            prediction = {"image_id": input["image_id"]}

            # TODO this is ugly
            if "instances" in output:
                instances = output["instances"].to(self._cpu_device)
                prediction["instances"] = instances_to_coco_json(instances, input["image_id"])
            if "proposals" in output:
                prediction["proposals"] = output["proposals"].to(self._cpu_device)
            self._predictions.append(prediction)

    def evaluate(self):
        """Evaluate"""
        if self._distributed:
            comm.synchronize()
            predictions = comm.gather(self._predictions, dst=0)
            predictions = list(itertools.chain(*predictions))

            if not comm.is_main_process():
                return {}
        else:
            predictions = self._predictions

        if len(predictions) == 0:
            self._logger.warning("[COCOEvaluator] Did not receive valid predictions.")
            return {}

        if self._output_dir:
            os.makedirs(self._output_dir, exist_ok=True)
            file_path = os.path.join(self._output_dir, "instances_predictions.pth")
            with open(file_path, "wb") as f:
                torch.save(predictions, f)

        self._results = OrderedDict()
        if "proposals" in predictions[0]:
            self._eval_box_proposals(predictions)
        if "instances" in predictions[0]:
            self._eval_predictions(set(self._tasks), predictions)
        # Copy so the caller can do whatever with results
        return copy.deepcopy(self._results)

    def _eval_predictions(self, tasks, predictions):
        """
        Evaluate predictions on the given tasks.
        Fill self._results with the metrics of the tasks.
        """
        self._logger.info("Preparing results for COCO format ...")
        coco_results = list(itertools.chain(*[x["instances"] for x in predictions]))

        # unmap the category ids for COCO
        if hasattr(self._metadata, "thing_dataset_id_to_contiguous_id"):
            reverse_id_mapping = {
                v: k for k, v in self._metadata.thing_dataset_id_to_contiguous_id.items()
            }
            for result in coco_results:
                category_id = result["category_id"]
                assert (
                    category_id in reverse_id_mapping
                ), "A prediction has category_id={}, which is not available in the dataset.".format(
                    category_id
                )
                result["category_id"] = reverse_id_mapping[category_id]

        if self._output_dir:
            file_path = os.path.join(self._output_dir, "coco_instances_results.json")
            self._logger.info("Saving results to {}".format(file_path))
            with open(file_path, "w") as f:
                f.write(json.dumps(coco_results))
                f.flush()

        if not self._do_evaluation:
            self._logger.info("Annotations are not available for evaluation.")
            return

        self._logger.info("Evaluating predictions ...")
        for task in sorted(tasks):
            coco_eval = (
                _evaluate_predictions_on_coco(
                    self._coco_api, coco_results, task, kpt_oks_sigmas=self._kpt_oks_sigmas
                )
                if len(coco_results) > 0
                else None  # cocoapi does not handle empty results very well
            )

            res = self._derive_coco_results(
                coco_eval, task, class_names=self._metadata.get("thing_classes")
            )
            self._results[task] = res

    def _eval_box_proposals(self, predictions):
        """
        Evaluate the box proposals in predictions.
        Fill self._results with the metrics for "box_proposals" task.
        """
        if self._output_dir:
            # Saving generated box proposals to file.
            # Predicted box_proposals are in XYXY_ABS mode.
            bbox_mode = BoxMode.XYXY_ABS.value
            ids, boxes, objectness_logits = [], [], []
            for prediction in predictions:
                ids.append(prediction["image_id"])
                boxes.append(prediction["proposals"].proposal_boxes.tensor.numpy())
                objectness_logits.append(prediction["proposals"].objectness_logits.numpy())

            proposal_data = {
                "boxes": boxes,
                "objectness_logits": objectness_logits,
                "ids": ids,
                "bbox_mode": bbox_mode,
            }
            with open(os.path.join(self._output_dir, "box_proposals.pkl"), "wb") as f:
                pickle.dump(proposal_data, f)

        if not self._do_evaluation:
            self._logger.info("Annotations are not available for evaluation.")
            return

        self._logger.info("Evaluating bbox proposals ...")
        res = {}
        areas = {"all": "", "small": "s", "medium": "m", "large": "l"}
        for limit in [100, 1000]:
            for area, suffix in areas.items():
                stats = _evaluate_box_proposals(predictions, self._coco_api, area=area, limit=limit)
                key = "AR{}@{:d}".format(suffix, limit)
                res[key] = float(stats["ar"].item() * 100)
        self._logger.info("Proposal metrics: \n" + pd.DataFrame([res.values()], columns=res.keys()).to_string(index=False))
        self._results["box_proposals"] = res

    def _derive_coco_results(self, coco_eval, iou_type, class_names=None):
        """
        Derive the desired score numbers from summarized COCOeval.

        Args:
            coco_eval (None or COCOEval): None represents no predictions from model.
            iou_type (str):
            class_names (None or list[str]): if provided, will use it to predict
                per-category AP.

        Returns:
            a dict of {metric name: score}
        """

        metrics = {
            "bbox": ["AP", "AP50", "AP75", "APs", "APm", "APl"],
            "segm": ["AP", "AP50", "AP75", "APs", "APm", "APl"],
            "keypoints": ["AP", "AP50", "AP75", "APm", "APl"],
        }[iou_type]

        if coco_eval is None:
            self._logger.warning("No predictions from the model!")
            return {metric: float("nan") for metric in metrics}

        # the standard metrics
        results = {
            metric: float(coco_eval.stats[idx] * 100 if coco_eval.stats[idx] >= 0 else "nan")
            for idx, metric in enumerate(metrics)
        }
        self._logger.info(
            "Evaluation results for {}: \n".format(iou_type) + pd.DataFrame([results.values()], columns=results.keys()).to_string(index=False)
        )
        if not np.isfinite(sum(results.values())):
            self._logger.info("Some metrics cannot be computed and is shown as NaN.")

        if class_names is None or len(class_names) <= 1:
            return results
        # Compute per-category AP
        # from https://github.com/facebookresearch/Detectron/blob/a6a835f5b8208c45d0dce217ce9bbda915f44df7/detectron/datasets/json_dataset_evaluator.py#L222-L252 # noqa
        precisions = coco_eval.eval["precision"]
        # precision has dims (iou, recall, cls, area range, max dets)
        assert len(class_names) == precisions.shape[2]

        results_per_category = []
        for idx, name in enumerate(class_names):
            # area range index 0: all area ranges
            # max dets index -1: typically 100 per image
            precision = precisions[:, :, idx, 0, -1]
            precision = precision[precision > -1]
            ap = np.mean(precision) if precision.size else float("nan")
            results_per_category.append(("{}".format(name), float(ap * 100)))

        table = pd.DataFrame(results_per_category, columns=('category', 'ap')).to_string(index=False)
        self._logger.info("Per-category {} AP: \n".format(iou_type) + table)

        results.update({"AP-" + name: ap for name, ap in results_per_category})
        return results


def instances_to_coco_json(instances, img_id):
    """
    Dump an "Instances" object to a COCO-format json that's used for evaluation.

    Args:
        instances (Instances):
        img_id (int): the image id

    Returns:
        list[dict]: list of json annotations in COCO format.
    """
    try_import_pycocotools()
    import pycocotools.mask as mask_util
    num_instance = len(instances)
    if num_instance == 0:
        return []

    boxes = instances.pred_boxes.tensor.numpy()
    boxes = BoxMode.convert(boxes, BoxMode.XYXY_ABS, BoxMode.XYWH_ABS)
    boxes = boxes.tolist()
    scores = instances.scores.tolist()
    classes = instances.pred_classes.tolist()

    has_mask = instances.has("pred_masks")
    if has_mask:
        # use RLE to encode the masks, because they are too large and takes memory
        # since this evaluator stores outputs of the entire dataset
        rles = [
            mask_util.encode(np.array(mask[:, :, None], order="F", dtype="uint8"))[0]
            for mask in instances.pred_masks
        ]
        for rle in rles:
            # "counts" is an array encoded by mask_util as a byte-stream. Python3's
            # json writer which always produces strings cannot serialize a bytestream
            # unless you decode it. Thankfully, utf-8 works out (which is also what
            # the pycocotools/_mask.pyx does).
            rle["counts"] = rle["counts"].decode("utf-8")

    has_keypoints = instances.has("pred_keypoints")
    if has_keypoints:
        keypoints = instances.pred_keypoints

    results = []
    for k in range(num_instance):
        result = {
            "image_id": img_id,
            "category_id": classes[k],
            "bbox": boxes[k],
            "score": scores[k],
        }
        if has_mask:
            result["segmentation"] = rles[k]
        if has_keypoints:
            # In COCO annotations,
            # keypoints coordinates are pixel indices.
            # However our predictions are floating point coordinates.
            # Therefore we subtract 0.5 to be consistent with the annotation format.
            # This is the inverse of data loading logic in `datasets/coco.py`.
            keypoints[k][:, :2] -= 0.5
            result["keypoints"] = keypoints[k].flatten().tolist()
        results.append(result)
    return results


# inspired from Detectron:
# https://github.com/facebookresearch/Detectron/blob/a6a835f5b8208c45d0dce217ce9bbda915f44df7/detectron/datasets/json_dataset_evaluator.py#L255 # noqa
def _evaluate_box_proposals(dataset_predictions, coco_api, thresholds=None, area="all", limit=None):
    """
    Evaluate detection proposal recall metrics. This function is a much
    faster alternative to the official COCO API recall evaluation code. However,
    it produces slightly different results.
    """
    # Record max overlap value for each gt box
    # Return vector of overlap values
    areas = {
        "all": 0,
        "small": 1,
        "medium": 2,
        "large": 3,
        "96-128": 4,
        "128-256": 5,
        "256-512": 6,
        "512-inf": 7,
    }
    area_ranges = [
        [0 ** 2, 1e5 ** 2],  # all
        [0 ** 2, 32 ** 2],  # small
        [32 ** 2, 96 ** 2],  # medium
        [96 ** 2, 1e5 ** 2],  # large
        [96 ** 2, 128 ** 2],  # 96-128
        [128 ** 2, 256 ** 2],  # 128-256
        [256 ** 2, 512 ** 2],  # 256-512
        [512 ** 2, 1e5 ** 2],
    ]  # 512-inf
    assert area in areas, "Unknown area range: {}".format(area)
    area_range = area_ranges[areas[area]]
    gt_overlaps = []
    num_pos = 0

    for prediction_dict in dataset_predictions:
        predictions = prediction_dict["proposals"]

        # sort predictions in descending order
        # TODO maybe remove this and make it explicit in the documentation
        inds = predictions.objectness_logits.sort(descending=True)[1]
        predictions = predictions[inds]

        ann_ids = coco_api.getAnnIds(imgIds=prediction_dict["image_id"])
        anno = coco_api.loadAnns(ann_ids)
        gt_boxes = [
            BoxMode.convert(obj["bbox"], BoxMode.XYWH_ABS, BoxMode.XYXY_ABS)
            for obj in anno
            if obj["iscrowd"] == 0
        ]
        gt_boxes = torch.as_tensor(gt_boxes).reshape(-1, 4)  # guard against no boxes
        gt_boxes = Boxes(gt_boxes)
        gt_areas = torch.as_tensor([obj["area"] for obj in anno if obj["iscrowd"] == 0])

        if len(gt_boxes) == 0 or len(predictions) == 0:
            continue

        valid_gt_inds = (gt_areas >= area_range[0]) & (gt_areas <= area_range[1])
        gt_boxes = gt_boxes[valid_gt_inds]

        num_pos += len(gt_boxes)

        if len(gt_boxes) == 0:
            continue

        if limit is not None and len(predictions) > limit:
            predictions = predictions[:limit]

        overlaps = pairwise_iou(predictions.proposal_boxes, gt_boxes)

        _gt_overlaps = torch.zeros(len(gt_boxes))
        for j in range(min(len(predictions), len(gt_boxes))):
            # find which proposal box maximally covers each gt box
            # and get the iou amount of coverage for each gt box
            max_overlaps, argmax_overlaps = overlaps.max(dim=0)

            # find which gt box is 'best' covered (i.e. 'best' = most iou)
            gt_ovr, gt_ind = max_overlaps.max(dim=0)
            assert gt_ovr >= 0
            # find the proposal box that covers the best covered gt box
            box_ind = argmax_overlaps[gt_ind]
            # record the iou coverage of this gt box
            _gt_overlaps[j] = overlaps[box_ind, gt_ind]
            assert _gt_overlaps[j] == gt_ovr
            # mark the proposal box and the gt box as used
            overlaps[box_ind, :] = -1
            overlaps[:, gt_ind] = -1

        # append recorded iou coverage level
        gt_overlaps.append(_gt_overlaps)
    gt_overlaps = (
        torch.cat(gt_overlaps, dim=0) if len(gt_overlaps) else torch.zeros(0, dtype=torch.float32)
    )
    gt_overlaps, _ = torch.sort(gt_overlaps)

    if thresholds is None:
        step = 0.05
        thresholds = torch.arange(0.5, 0.95 + 1e-5, step, dtype=torch.float32)
    recalls = torch.zeros_like(thresholds)
    # compute recall for each iou threshold
    for i, t in enumerate(thresholds):
        recalls[i] = (gt_overlaps >= t).float().sum() / float(num_pos)
    # ar = 2 * np.trapz(recalls, thresholds)
    ar = recalls.mean()
    return {
        "ar": ar,
        "recalls": recalls,
        "thresholds": thresholds,
        "gt_overlaps": gt_overlaps,
        "num_pos": num_pos,
    }


def _evaluate_predictions_on_coco(coco_gt, coco_results, iou_type, kpt_oks_sigmas=None):
    """
    Evaluate the coco results using COCOEval API.
    """
    try_import_pycocotools()
    from pycocotools.cocoeval import COCOeval
    assert len(coco_results) > 0

    if iou_type == "segm":
        coco_results = copy.deepcopy(coco_results)
        # When evaluating mask AP, if the results contain bbox, cocoapi will
        # use the box area as the area of the instance, instead of the mask area.
        # This leads to a different definition of small/medium/large.
        # We remove the bbox field to let mask AP use mask area.
        for c in coco_results:
            c.pop("bbox", None)

    coco_dt = coco_gt.loadRes(coco_results)
    coco_eval = COCOeval(coco_gt, coco_dt, iou_type)

    if iou_type == "keypoints":
        # Use the COCO default keypoint OKS sigmas unless overrides are specified
        if kpt_oks_sigmas:
            assert hasattr(coco_eval.params, "kpt_oks_sigmas"), "pycocotools is too old!"
            coco_eval.params.kpt_oks_sigmas = np.array(kpt_oks_sigmas)
        # COCOAPI requires every detection and every gt to have keypoints, so
        # we just take the first entry from both
        num_keypoints_dt = len(coco_results[0]["keypoints"]) // 3
        num_keypoints_gt = len(next(iter(coco_gt.anns.values()))["keypoints"]) // 3
        num_keypoints_oks = len(coco_eval.params.kpt_oks_sigmas)
        assert num_keypoints_oks == num_keypoints_dt == num_keypoints_gt, (
            f"[COCOEvaluator] Prediction contain {num_keypoints_dt} keypoints. "
            f"Ground truth contains {num_keypoints_gt} keypoints. "
            f"The length of cfg.CONFIG.VAL.KEYPOINT_OKS_SIGMAS is {num_keypoints_oks}. "
            "They have to agree with each other. For meaning of OKS, please refer to "
            "http://cocodataset.org/#keypoints-eval."
        )

    coco_eval.evaluate()
    coco_eval.accumulate()
    coco_eval.summarize()

    return coco_eval



"""Directpose based on resnet + FPN backbones"""
# pylint: disable=line-too-long, redefined-builtin, missing-class-docstring, unused-variable,unnecessary-pass,arguments-differ
# adpated from https://github.com/aim-uofa/AdelaiDet/blob/master/adet/modeling/backbone
import math
from abc import ABCMeta, abstractmethod

import numpy as np
import torch
import torch.nn.parallel
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo

from ...data.structures import ImageList
from ...engine.config import get_cfg_defaults
from ...nn.batch_norm import get_norm
from ...nn.batch_norm import NaiveSyncBatchNorm
from ...nn.shape_spec import ShapeSpec
from ..object_detection.model_utils import detector_postprocess
from .directpose import DirectPose


__all__ = ['directpose_resnet_lpf_fpn', 'directpose_resnet50_lpf_fpn_coco']


def conv3x3(in_planes, out_planes, stride=1, groups=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, groups=groups, bias=False)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class Conv2d(torch.nn.Conv2d):
    """
    A wrapper around :class:`torch.nn.Conv2d` to support empty inputs and more features.
    """

    def __init__(self, *args, **kwargs):
        """
        Extra keyword arguments supported in addition to those in `torch.nn.Conv2d`:
        Args:
            norm (nn.Module, optional): a normalization layer
            activation (callable(Tensor) -> Tensor): a callable activation function
        It assumes that norm layer is used before activation.
        """
        norm = kwargs.pop("norm", None)
        activation = kwargs.pop("activation", None)
        super().__init__(*args, **kwargs)

        self.norm = norm
        self.activation = activation

    def forward(self, x):
        if x.numel() == 0 and self.training:
            # https://github.com/pytorch/pytorch/issues/12013
            assert not isinstance(
                self.norm, torch.nn.SyncBatchNorm
            ), "SyncBatchNorm does not support empty inputs!"

        x = super().forward(x)
        if self.norm is not None:
            x = self.norm(x)
        if self.activation is not None:
            x = self.activation(x)
        return x


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1, norm_layer=None, filter_size=1):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1:
            raise ValueError('BasicBlock only supports groups=1')
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        if(stride == 1):
            self.conv2 = conv3x3(planes, planes)
        else:
            self.conv2 = nn.Sequential(Downsample(filt_size=filter_size, stride=stride, channels=planes),
                                       conv3x3(planes, planes),)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1, norm_layer=None, filter_size=1):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, planes)
        self.bn1 = norm_layer(planes)
        self.conv2 = conv3x3(planes, planes, groups)  # stride moved
        self.bn2 = norm_layer(planes)
        if(stride == 1):
            self.conv3 = conv1x1(planes, planes * self.expansion)
        else:
            self.conv3 = nn.Sequential(Downsample(filt_size=filter_size, stride=stride, channels=planes),
                                       conv1x1(planes, planes * self.expansion))
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Backbone(nn.Module, metaclass=ABCMeta):
    """
    Abstract base class for network backbones.
    """

    def __init__(self):
        """
        The `__init__` method of any subclass can specify its own set of arguments.
        """
        super().__init__()

    @abstractmethod
    def forward(self):
        """
        Subclasses must override this method, but adhere to the same return type.
        Returns:
            dict[str->Tensor]: mapping from feature name (e.g., "res2") to tensor
        """
        pass

    @property
    def size_divisibility(self):
        """
        Some backbones require the input height and width to be divisible by a
        specific integer. This is typically true for encoder / decoder type networks
        with lateral connection (e.g., FPN) for which feature maps need to match
        dimension in the "bottom up" and "top down" paths. Set to 0 if no specific
        input size divisibility is required.
        """
        return 0

    def output_shape(self):
        """
        Returns:
            dict[str->ShapeSpec]
        """
        # this is a backward-compatible default
        return {
            name: ShapeSpec(
                channels=self._out_feature_channels[name], stride=self._out_feature_strides[name]
            )
            for name in self._out_features
        }


class ResNetLPF(Backbone):
    def __init__(self, cfg, block, layers, zero_init_residual=False,
                 groups=1, width_per_group=64, norm_layer=None, filter_size=1,
                 pool_only=True, return_idx=(0, 1, 2, 3)):
        super().__init__()
        self.return_idx = list(return_idx)
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        planes = [int(width_per_group * groups * 2 ** i) for i in range(4)]
        self.inplanes = planes[0]

        if(pool_only):
            self.conv1 = nn.Conv2d(3, planes[0], kernel_size=7, stride=2, padding=3, bias=False)
        else:
            self.conv1 = nn.Conv2d(3, planes[0], kernel_size=7, stride=1, padding=3, bias=False)
        self.bn1 = norm_layer(planes[0])
        self.relu = nn.ReLU(inplace=True)

        if(pool_only):
            self.maxpool = nn.Sequential(*[nn.MaxPool2d(kernel_size=2, stride=1),
                                           Downsample(filt_size=filter_size, stride=2, channels=planes[0])])
        else:
            self.maxpool = nn.Sequential(*[Downsample(filt_size=filter_size, stride=2, channels=planes[0]),
                                           nn.MaxPool2d(kernel_size=2, stride=1),
                                           Downsample(filt_size=filter_size, stride=2, channels=planes[0])])

        self.layer1 = self._make_layer(
            block, planes[0], layers[0], groups=groups, norm_layer=norm_layer)
        self.layer2 = self._make_layer(
            block, planes[1], layers[1], stride=2, groups=groups, norm_layer=norm_layer, filter_size=filter_size)
        self.layer3 = self._make_layer(
            block, planes[2], layers[2], stride=2, groups=groups, norm_layer=norm_layer, filter_size=filter_size)
        self.layer4 = self._make_layer(
            block, planes[3], layers[3], stride=2, groups=groups, norm_layer=norm_layer, filter_size=filter_size)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if(m.in_channels != m.out_channels or m.out_channels != m.groups or m.bias is not None):
                    # don't want to reinitialize downsample layers, code assuming normal conv layers will not have these characteristics
                    nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                else:
                    print('Not initializing')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)
        self._freeze_backbone(cfg.CONFIG.MODEL.BACKBONE.FREEZE_AT)
        # self._freeze_bn()

    def _freeze_backbone(self, freeze_at):
        if freeze_at < 0:
            return
        for stage_index in range(freeze_at):
            if stage_index == 0:
                # stage 0 is the stem
                for p in self.conv1.parameters():
                    p.requires_grad = False
                for p in self.bn1.parameters():
                    p.requires_grad = False
            else:
                m = getattr(self, "layer" + str(stage_index))
                for p in m.parameters():
                    p.requires_grad = False

    def _freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def _make_layer(self, block, planes, blocks, stride=1, groups=1, norm_layer=None, filter_size=1):
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            # downsample = nn.Sequential(
            #     conv1x1(self.inplanes, planes * block.expansion, stride, filter_size=filter_size),
            #     norm_layer(planes * block.expansion),
            # )

            downsample = [Downsample(filt_size=filter_size, stride=stride,
                                     channels=self.inplanes), ] if(stride != 1) else []
            downsample += [conv1x1(self.inplanes, planes * block.expansion, 1),
                           norm_layer(planes * block.expansion)]
            # print(downsample)
            downsample = nn.Sequential(*downsample)

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample,
                            groups, norm_layer, filter_size=filter_size))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=groups,
                                norm_layer=norm_layer, filter_size=filter_size))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        outs = []
        outs.append(self.layer1(x))  # 1/4
        outs.append(self.layer2(outs[-1]))  # 1/8
        outs.append(self.layer3(outs[-1]))  # 1/16
        outs.append(self.layer4(outs[-1]))  # 1/32
        return {"res{}".format(idx + 2): outs[idx] for idx in self.return_idx}


class Downsample(nn.Module):
    def __init__(self, pad_type='reflect', filt_size=3, stride=2, channels=None, pad_off=0):
        super(Downsample, self).__init__()
        self.filt_size = filt_size
        self.pad_off = pad_off
        self.pad_sizes = [int(1.*(filt_size-1)/2), int(np.ceil(1.*(filt_size-1)/2)), int(1.*(filt_size-1)/2), int(np.ceil(1.*(filt_size-1)/2))]
        self.pad_sizes = [pad_size+pad_off for pad_size in self.pad_sizes]
        self.stride = stride
        self.off = int((self.stride-1)/2.)
        self.channels = channels

        # print('Filter size [%i]'%filt_size)
        if(self.filt_size == 1):
            a = np.array([1.,])
        elif(self.filt_size == 2):
            a = np.array([1., 1.])
        elif(self.filt_size == 3):
            a = np.array([1., 2., 1.])
        elif(self.filt_size == 4):
            a = np.array([1., 3., 3., 1.])
        elif(self.filt_size == 5):
            a = np.array([1., 4., 6., 4., 1.])
        elif(self.filt_size == 6):
            a = np.array([1., 5., 10., 10., 5., 1.])
        elif(self.filt_size == 7):
            a = np.array([1., 6., 15., 20., 15., 6., 1.])

        filt = torch.Tensor(a[:, None] * a[None, :])
        filt = filt / torch.sum(filt)
        self.register_buffer('filt', filt[None, None, :, :].repeat((self.channels, 1, 1, 1)))

        self.pad = get_pad_layer(pad_type)(self.pad_sizes)

    def forward(self, inp):
        if(self.filt_size == 1):
            if(self.pad_off == 0):
                return inp[:, :, ::self.stride, ::self.stride]
            else:
                return self.pad(inp)[:, :, ::self.stride, ::self.stride]
        else:
            return F.conv2d(self.pad(inp), self.filt, stride=self.stride, groups=inp.shape[1])

def get_pad_layer(pad_type):
    if(pad_type in ['refl', 'reflect']):
        PadLayer = nn.ReflectionPad2d
    elif(pad_type in ['repl', 'replicate']):
        PadLayer = nn.ReplicationPad2d
    elif(pad_type == 'zero'):
        PadLayer = nn.ZeroPad2d
    else:
        print('Pad type [%s] not recognized'%pad_type)
    return PadLayer


class Downsample1D(nn.Module):
    def __init__(self, pad_type='reflect', filt_size=3, stride=2, channels=None, pad_off=0):
        super(Downsample1D, self).__init__()
        self.filt_size = filt_size
        self.pad_off = pad_off
        self.pad_sizes = [int(1. * (filt_size - 1) / 2), int(np.ceil(1. * (filt_size - 1) / 2))]
        self.pad_sizes = [pad_size + pad_off for pad_size in self.pad_sizes]
        self.stride = stride
        self.off = int((self.stride - 1) / 2.)
        self.channels = channels

        # print('Filter size [%i]' % filt_size)
        if(self.filt_size == 1):
            a = np.array([1., ])
        elif(self.filt_size == 2):
            a = np.array([1., 1.])
        elif(self.filt_size == 3):
            a = np.array([1., 2., 1.])
        elif(self.filt_size == 4):
            a = np.array([1., 3., 3., 1.])
        elif(self.filt_size == 5):
            a = np.array([1., 4., 6., 4., 1.])
        elif(self.filt_size == 6):
            a = np.array([1., 5., 10., 10., 5., 1.])
        elif(self.filt_size == 7):
            a = np.array([1., 6., 15., 20., 15., 6., 1.])

        filt = torch.Tensor(a)
        filt = filt / torch.sum(filt)
        self.register_buffer('filt', filt[None, None, :].repeat((self.channels, 1, 1)))

        self.pad = get_pad_layer_1d(pad_type)(self.pad_sizes)

    def forward(self, inp):
        if(self.filt_size == 1):
            if(self.pad_off == 0):
                return inp[:, :, ::self.stride]
            else:
                return self.pad(inp)[:, :, ::self.stride]
        else:
            return F.conv1d(self.pad(inp), self.filt, stride=self.stride, groups=inp.shape[1])


def get_pad_layer_1d(pad_type):
    if(pad_type in ['refl', 'reflect']):
        PadLayer = nn.ReflectionPad1d
    elif(pad_type in ['repl', 'replicate']):
        PadLayer = nn.ReplicationPad1d
    elif(pad_type == 'zero'):
        PadLayer = nn.ZeroPad1d
    else:
        print('Pad type [%s] not recognized' % pad_type)
    return PadLayer


class FPN(Backbone):
    """
    This module implements :paper:`FPN`.
    It creates pyramid features built on top of some input feature maps.
    """

    def __init__(self, bottom_up, in_features, out_channels, norm="", top_block=None, fuse_type="sum"):
        """
        Args:
            bottom_up (Backbone): module representing the bottom up subnetwork.
                Must be a subclass of :class:`Backbone`. The multi-scale feature
                maps generated by the bottom up network, and listed in `in_features`,
                are used to generate FPN levels.
            in_features (list[str]): names of the input feature maps coming
                from the backbone to which FPN is attached. For example, if the
                backbone produces ["res2", "res3", "res4"], any *contiguous* sublist
                of these may be used; order must be from high to low resolution.
            out_channels (int): number of channels in the output feature maps.
            norm (str): the normalization to use.
            top_block (nn.Module or None): if provided, an extra operation will
                be performed on the output of the last (smallest resolution)
                FPN output, and the result will extend the result list. The top_block
                further downsamples the feature map. It must have an attribute
                "num_levels", meaning the number of extra FPN levels added by
                this block, and "in_feature", which is a string representing
                its input feature (e.g., p5).
            fuse_type (str): types for fusing the top down features and the lateral
                ones. It can be "sum" (default), which sums up element-wise; or "avg",
                which takes the element-wise mean of the two.
        """
        super(FPN, self).__init__()
        assert isinstance(bottom_up, Backbone)

        # Feature map strides and channels from the bottom up network (e.g. ResNet)
        input_shapes = bottom_up.output_shape()
        in_strides = [input_shapes[f].stride for f in in_features]
        in_channels = [input_shapes[f].channels for f in in_features]

        _assert_strides_are_log2_contiguous(in_strides)
        lateral_convs = []
        output_convs = []

        use_bias = norm == ""
        for idx, in_channels in enumerate(in_channels):
            lateral_norm = get_norm(norm, out_channels)
            output_norm = get_norm(norm, out_channels)

            lateral_conv = Conv2d(
                in_channels, out_channels, kernel_size=1, bias=use_bias, norm=lateral_norm
            )
            output_conv = Conv2d(
                out_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=use_bias,
                norm=output_norm,
            )
            c2_xavier_fill(lateral_conv)
            c2_xavier_fill(output_conv)
            stage = int(math.log2(in_strides[idx]))
            self.add_module("fpn_lateral{}".format(stage), lateral_conv)
            self.add_module("fpn_output{}".format(stage), output_conv)

            lateral_convs.append(lateral_conv)
            output_convs.append(output_conv)
        # Place convs into top-down order (from low to high resolution)
        # to make the top-down computation in forward clearer.
        self.lateral_convs = lateral_convs[::-1]
        self.output_convs = output_convs[::-1]
        self.top_block = top_block
        self.in_features = in_features
        self.bottom_up = bottom_up
        # Return feature names are "p<stage>", like ["p2", "p3", ..., "p6"]
        self._out_feature_strides = {"p{}".format(int(math.log2(s))): s for s in in_strides}
        # top block output feature maps.
        if self.top_block is not None:
            for s in range(stage, stage + self.top_block.num_levels):
                self._out_feature_strides["p{}".format(s + 1)] = 2 ** (s + 1)

        self._out_features = list(self._out_feature_strides.keys())
        self._out_feature_channels = {k: out_channels for k in self._out_features}
        self._size_divisibility = in_strides[-1]
        assert fuse_type in {"avg", "sum"}
        self._fuse_type = fuse_type

    @property
    def size_divisibility(self):
        return self._size_divisibility

    def forward(self, x):
        """
        Args:
            input (dict[str->Tensor]): mapping feature map name (e.g., "res5") to
                feature map tensor for each feature level in high to low resolution order.
        Returns:
            dict[str->Tensor]:
                mapping from feature map name to FPN feature map tensor
                in high to low resolution order. Returned feature names follow the FPN
                paper convention: "p<stage>", where stage has stride = 2 ** stage e.g.,
                ["p2", "p3", ..., "p6"].
        """
        # Reverse feature maps into top-down order (from low to high resolution)
        # print(x.shape, x[:, :, 100:200, 100:200])
        # raise
        bottom_up_features = self.bottom_up(x)
        x = [bottom_up_features[f] for f in self.in_features[::-1]]
        results = []
        prev_features = self.lateral_convs[0](x[0])
        results.append(self.output_convs[0](prev_features))
        for features, lateral_conv, output_conv in zip(x[1:], self.lateral_convs[1:], self.output_convs[1:]):
            lateral_features = lateral_conv(features)
            top_down_features = F.interpolate(prev_features, size=lateral_features.shape[-2:], mode="nearest")
            prev_features = lateral_features + top_down_features
            if self._fuse_type == "avg":
                prev_features /= 2
            results.insert(0, output_conv(prev_features))

        if self.top_block is not None:
            top_block_in_feature = bottom_up_features.get(self.top_block.in_feature, None)
            if top_block_in_feature is None:
                top_block_in_feature = results[self._out_features.index(self.top_block.in_feature)]
            results.extend(self.top_block(top_block_in_feature))
        assert len(self._out_features) == len(results)
        return dict(zip(self._out_features, results))

    def output_shape(self):
        return {
            name: ShapeSpec(
                channels=self._out_feature_channels[name], stride=self._out_feature_strides[name]
            )
            for name in self._out_features
        }

def c2_xavier_fill(module: nn.Module) -> None:
    """
    Initialize `module.weight` using the "XavierFill" implemented in Caffe2.
    Also initializes `module.bias` to 0.
    Args:
        module (torch.nn.Module): module to initialize.
    """
    # Caffe2 implementation of XavierFill in fact
    # corresponds to kaiming_uniform_ in PyTorch
    nn.init.kaiming_uniform_(module.weight, a=1)  # pyre-ignore
    if module.bias is not None:  # pyre-ignore
        nn.init.constant_(module.bias, 0)

def _assert_strides_are_log2_contiguous(strides):
    """
    Assert that each stride is 2x times its preceding stride, i.e. "contiguous in log2".
    """
    for i, stride in enumerate(strides[1:], 1):
        assert stride == 2 * strides[i - 1], "Strides {} {} are not log2 contiguous".format(
            stride, strides[i - 1]
        )

class LastLevelP6P7(nn.Module):
    """
    This module is used in RetinaNet and FCOS to generate extra layers, P6 and P7 from
    C5 or P5 feature.
    """

    def __init__(self, in_channels, out_channels, in_features="res5"):
        super().__init__()
        self.num_levels = 2
        self.in_feature = in_features
        self.p6 = nn.Conv2d(in_channels, out_channels, 3, 2, 1)
        self.p7 = nn.Conv2d(out_channels, out_channels, 3, 2, 1)
        for module in [self.p6, self.p7]:
            c2_xavier_fill(module)

    def forward(self, x):
        p6 = self.p6(x)
        p7 = self.p7(F.relu(p6))
        return [p6, p7]


class LastLevelP6(nn.Module):
    """
    This module is used in FCOS to generate extra layers
    """

    def __init__(self, in_channels, out_channels, in_features="res5"):
        super().__init__()
        self.num_levels = 1
        self.in_feature = in_features
        self.p6 = nn.Conv2d(in_channels, out_channels, 3, 2, 1)
        for module in [self.p6]:
            c2_xavier_fill(module)

    def forward(self, x):
        p6 = self.p6(x)
        return [p6]


class ResNetFPNDirectPose(nn.Module):
    def __init__(self, backbone, pose, pixel_mean=(0.485, 0.456, 0.406),
                 pixel_std=(0.229, 0.224, 0.225)):
        super(ResNetFPNDirectPose, self).__init__()
        self.backbone = backbone
        self.proposal_generator = pose
        self.register_buffer("pixel_mean", torch.Tensor(pixel_mean).view(-1, 1, 1))
        self.register_buffer("pixel_std", torch.Tensor(pixel_std).view(-1, 1, 1))

    @property
    def device(self):
        return self.pixel_mean.device

    def forward(self, batched_inputs_or_tensor):
        if isinstance(batched_inputs_or_tensor, torch.Tensor):
            images = images_tensor = batched_inputs_or_tensor
            is_tensor = True
        else:
            # preprocess
            images = self.preprocess_image(batched_inputs_or_tensor)
            images_tensor = images.tensor
            is_tensor = False

        features = self.backbone(images_tensor)

        if not is_tensor and "instances" in batched_inputs_or_tensor[0]:
            gt_instances = [x["instances"].to(self.device) for x in batched_inputs_or_tensor]
        elif not is_tensor and "targets" in batched_inputs_or_tensor[0]:
            gt_instances = [x["targets"].to(self.device) for x in batched_inputs_or_tensor]
        else:
            gt_instances = None
        proposals, proposal_losses = self.proposal_generator(images, features, gt_instances)
        if self.training:
            return proposal_losses

        if is_tensor:
            # raw tensor input, e.g., during tracing
            return proposals
        processed_results = []
        for results_per_image, input_per_image, image_size in zip(proposals, batched_inputs_or_tensor, images.image_sizes):
            height = input_per_image.get("height", image_size[0])
            width = input_per_image.get("width", image_size[1])
            r = detector_postprocess(results_per_image, height, width)
            processed_results.append({"instances": r})
        return processed_results


    def preprocess_image(self, batched_inputs):
        """
        Normalize, pad and batch the input images.
        """
        images = [x["image"].to(self.device) for x in batched_inputs]
        images = [(x / 255 - self.pixel_mean) / self.pixel_std for x in images]
        images = ImageList.from_tensors(images, self.backbone.size_divisibility)
        return images


def directpose_resnet_lpf_fpn(cfg):
    """
    Create a ResNetFPNDirectPose instance from config.
    Returns:
        ResNetFPNDirectPose: a :class:`ResNetFPNDirectPose` instance.
    """
    depth = cfg.CONFIG.MODEL.RESNETS.DEPTH
    out_features = cfg.CONFIG.MODEL.RESNETS.OUT_FEATURES

    num_blocks_per_stage = {50: [3, 4, 6, 3], 101: [3, 4, 23, 3], 152: [3, 8, 36, 3]}[depth]
    out_stage_idx = [{"res2": 0, "res3": 1, "res4": 2, "res5": 3}[f] for f in out_features]
    out_feature_channels = {"res2": 256, "res3": 512,
                            "res4": 1024, "res5": 2048}
    out_feature_strides = {"res2": 4, "res3": 8, "res4": 16, "res5": 32}
    model = ResNetLPF(cfg, Bottleneck, num_blocks_per_stage, norm_layer=NaiveSyncBatchNorm,
                      filter_size=3, pool_only=True, return_idx=out_stage_idx)
    model._out_features = out_features
    model._out_feature_channels = out_feature_channels
    model._out_feature_strides = out_feature_strides

    if cfg.CONFIG.MODEL.BACKBONE.LOAD_URL:
        model.load_state_dict(
            model_zoo.load_url(cfg.CONFIG.MODEL.BACKBONE.LOAD_URL, map_location='cpu', check_hash=True)['state_dict'],
            strict=False)

    if cfg.CONFIG.MODEL.BACKBONE.ANTI_ALIAS:
        bottom_up = model
    else:
        raise NotImplementedError
    in_features = cfg.CONFIG.MODEL.FPN.IN_FEATURES
    out_channels = cfg.CONFIG.MODEL.FPN.OUT_CHANNELS
    top_levels = cfg.CONFIG.MODEL.DIRECTPOSE.TOP_LEVELS
    in_channels_top = out_channels
    if top_levels == 2:
        top_block = LastLevelP6P7(in_channels_top, out_channels, "p5")
    if top_levels == 1:
        top_block = LastLevelP6(in_channels_top, out_channels, "p5")
    elif top_levels == 0:
        top_block = None
    backbone = FPN(
        bottom_up=bottom_up,
        in_features=in_features,
        out_channels=out_channels,
        norm=cfg.CONFIG.MODEL.FPN.NORM,
        top_block=top_block,
        fuse_type=cfg.CONFIG.MODEL.FPN.FUSE_TYPE,
    )

    pose_net = DirectPose(cfg, backbone.output_shape())

    return ResNetFPNDirectPose(backbone, pose_net)

def directpose_resnet50_lpf_fpn_coco(cfg):
    base_cfg = get_cfg_defaults("directpose")
    base_cfg.merge_from_other_cfg(cfg)
    cfg = base_cfg
    depth = 50
    out_features = ["res3", "res4", "res5"]

    num_blocks_per_stage = {50: [3, 4, 6, 3], 101: [3, 4, 23, 3], 152: [3, 8, 36, 3]}[depth]
    out_stage_idx = [{"res2": 0, "res3": 1, "res4": 2, "res5": 3}[f] for f in out_features]
    out_feature_channels = {"res2": 256, "res3": 512,
                            "res4": 1024, "res5": 2048}
    out_feature_strides = {"res2": 4, "res3": 8, "res4": 16, "res5": 32}
    model = ResNetLPF(cfg, Bottleneck, num_blocks_per_stage, norm_layer=NaiveSyncBatchNorm,
                      filter_size=3, pool_only=True, return_idx=out_stage_idx)
    model._out_features = out_features
    model._out_feature_channels = out_feature_channels
    model._out_feature_strides = out_feature_strides

    if cfg.CONFIG.MODEL.BACKBONE.LOAD_URL and not cfg.CONFIG.MODEL.PRETRAINED:
        model.load_state_dict(
            model_zoo.load_url(cfg.CONFIG.MODEL.BACKBONE.LOAD_URL, map_location='cpu', check_hash=True)['state_dict'],
            strict=False)

    bottom_up = model
    in_features = ["res3", "res4", "res5"]
    out_channels = 256
    in_channels_top = out_channels
    top_block = LastLevelP6P7(in_channels_top, out_channels, "p5")
    backbone = FPN(
        bottom_up=bottom_up,
        in_features=in_features,
        out_channels=out_channels,
        norm="",
        top_block=top_block,
        fuse_type="sum",
    )

    cfg.CONFIG.MODEL.DIRECTPOSE.TOP_LEVELS = 2
    cfg.CONFIG.MODEL.DIRECTPOSE.SIZES_OF_INTEREST = [64, 128, 256, 512]
    cfg.CONFIG.MODEL.DIRECTPOSE.FPN_STRIDES = [8, 16, 32, 64, 128]
    cfg.CONFIG.MODEL.DIRECTPOSE.IN_FEATURES = ['p3', 'p4', 'p5', 'p6', 'p7']
    cfg.CONFIG.MODEL.DIRECTPOSE.HM_OFFSET = False
    cfg.CONFIG.MODEL.DIRECTPOSE.HM_TYPE = 'BinaryLabels'
    cfg.CONFIG.MODEL.DIRECTPOSE.REFINE_KPT = False
    cfg.CONFIG.MODEL.DIRECTPOSE.SAMPLE_FEATURE = 'lower'
    cfg.CONFIG.MODEL.DIRECTPOSE.KPALIGN_GROUPS = 9
    cfg.CONFIG.MODEL.DIRECTPOSE.SEPERATE_CONV_FEATURE = True
    cfg.CONFIG.MODEL.DIRECTPOSE.LOSS_ON_LOCATOR = False
    cfg.CONFIG.MODEL.DIRECTPOSE.KPT_VIS = True
    cfg.CONFIG.MODEL.DIRECTPOSE.CLOSEKPT_NMS = False
    cfg.CONFIG.MODEL.DIRECTPOSE.CENTER_BRANCH = 'kpt'
    cfg.CONFIG.MODEL.DIRECTPOSE.USE_SCALE = False

    pose_net = DirectPose(cfg, backbone.output_shape())

    model = ResNetFPNDirectPose(backbone, pose_net)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        state_dict = torch.load(get_model_file('directpose_resnet50_lpf_fpn_coco', tag=cfg.CONFIG.MODEL.PRETRAINED),
                                map_location=torch.device('cpu'))
        msg = model.load_state_dict(state_dict, strict=True)
        print("=> Initialized from a DirectPose pretrained on COCO dataset")
    return model



"""Region Proposal Networks Definition."""
from __future__ import absolute_import

import mxnet as mx
from mxnet import autograd
from mxnet import gluon
from mxnet.gluon import nn

from .anchor import RPNAnchorGenerator
from .proposal import RPNProposal


class RPN(gluon.HybridBlock):
    r"""Region Proposal Network.

    Parameters
    ----------
    channels : int
        Channel number used in convolutional layers.
    strides : int or tuple of ints
        Feature map stride with respect to original image.
        This is usually the ratio between original image size and feature map size.
    base_size : int
        The width(and height) of reference anchor box.
    scales : iterable of float
        The areas of anchor boxes.
        We use the following form to compute the shapes of anchors:

        .. math::

            width_{anchor} = size_{base} \times scale \times \sqrt{ 1 / ratio}
            height_{anchor} = size_{base} \times scale \times \sqrt{ratio}

    ratios : iterable of float
        The aspect ratios of anchor boxes. We expect it to be a list or tuple.
    alloc_size : tuple of int
        Allocate size for the anchor boxes as (H, W).
        Usually we generate enough anchors for large feature map, e.g. 128x128.
        Later in inference we can have variable input sizes,
        at which time we can crop corresponding anchors from this large
        anchor map so we can skip re-generating anchors for each input.
    clip : float
        Clip bounding box target to this value.
    nms_thresh : float
        IOU threshold for NMS. It is used to remove overlapping proposals.
    train_pre_nms : int
        Filter top proposals before NMS in training.
    train_post_nms : int
        Return top proposal results after NMS in training.
    test_pre_nms : int
        Filter top proposals before NMS in testing.
    test_post_nms : int
        Return top proposal results after NMS in testing.
    min_size : int
        Proposals whose size is smaller than ``min_size`` will be discarded.
        multi_level : boolean
        Whether to extract feature from multiple level. This is used in FPN.
    multi_level : boolean, default is False.
        Whether to use multiple feature maps for RPN. eg. FPN.
    per_level_nms : boollean, default is False
        Whether to apply nms on each level's rois instead of applying nms after aggregation.
    minimal_opset : bool, default is `False`
        We sometimes add special operators to accelerate training/inference, however, for exporting
        to third party compilers we want to utilize most widely used operators.
        If `minimal_opset` is `True`, the network will use a minimal set of operators good
        for e.g., `TVM`.
    """

    def __init__(self, channels, strides, base_size, scales, ratios, alloc_size,
                 clip, nms_thresh, train_pre_nms, train_post_nms,
                 test_pre_nms, test_post_nms, min_size, multi_level=False, per_level_nms=False,
                 minimal_opset=False, **kwargs):
        super(RPN, self).__init__(**kwargs)
        self._nms_thresh = nms_thresh
        self._multi_level = multi_level
        self._per_level_nms = per_level_nms
        self._train_pre_nms = max(1, train_pre_nms)
        self._train_post_nms = max(1, train_post_nms)
        self._test_pre_nms = max(1, test_pre_nms)
        self._test_post_nms = max(1, test_post_nms)
        num_stages = len(scales)
        with self.name_scope():
            if self._multi_level:
                asz = alloc_size
                self.anchor_generator = nn.HybridSequential()
                for _, st, s in zip(range(num_stages), strides, scales):
                    stage_anchor_generator = RPNAnchorGenerator(st, base_size, ratios, s, asz)
                    self.anchor_generator.add(stage_anchor_generator)
                    asz = max(asz[0] // 2, 16)
                    asz = (asz, asz)  # For FPN, We use large anchor presets
                anchor_depth = self.anchor_generator[0].num_depth
                self.rpn_head = RPNHead(channels, anchor_depth)
            else:
                self.anchor_generator = RPNAnchorGenerator(
                    strides, base_size, ratios, scales, alloc_size)
                anchor_depth = self.anchor_generator.num_depth
                # not using RPNHead to keep backward compatibility with old models
                weight_initializer = mx.init.Normal(0.01)
                self.conv1 = nn.HybridSequential()
                self.conv1.add(nn.Conv2D(channels, 3, 1, 1, weight_initializer=weight_initializer),
                               nn.Activation('relu'))
                self.score = nn.Conv2D(anchor_depth, 1, 1, 0, weight_initializer=weight_initializer)
                self.loc = nn.Conv2D(anchor_depth * 4, 1, 1, 0,
                                     weight_initializer=weight_initializer)

            self.region_proposer = RPNProposal(clip, min_size, stds=(1.0, 1.0, 1.0, 1.0), minimal_opset=minimal_opset)

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, img, *x):
        """Forward RPN.

        The behavior during training and inference is different.

        Parameters
        ----------
        img : mxnet.nd.NDArray or mxnet.symbol
            The original input image.
        x : mxnet.nd.NDArray or mxnet.symbol(s)
            Feature tensor(s).

        Returns
        -------
        (rpn_score, rpn_box)
            Returns predicted scores and regions which are candidates of objects.

        """
        if autograd.is_training():
            pre_nms = self._train_pre_nms
            post_nms = self._train_post_nms
        else:
            pre_nms = self._test_pre_nms
            post_nms = self._test_post_nms
        anchors = []
        rpn_pre_nms_proposals = []
        raw_rpn_scores = []
        raw_rpn_boxes = []
        if self._multi_level:
            # Generate anchors in [P2, P3, P4, P5, P6] order
            for i, feat in enumerate(x):
                ag = self.anchor_generator[i]
                anchor = ag(feat)
                rpn_score, rpn_box, raw_rpn_score, raw_rpn_box = \
                    self.rpn_head(feat)
                rpn_pre = self.region_proposer(anchor, rpn_score, rpn_box, img)
                if self._per_level_nms:
                    with autograd.pause():
                        # Non-maximum suppression
                        rpn_pre = F.contrib.box_nms(rpn_pre, overlap_thresh=self._nms_thresh,
                                                    topk=pre_nms // len(x), coord_start=1,
                                                    score_index=0, id_index=-1)
                anchors.append(anchor)
                rpn_pre_nms_proposals.append(rpn_pre)
                raw_rpn_scores.append(raw_rpn_score)
                raw_rpn_boxes.append(raw_rpn_box)
            rpn_pre_nms_proposals = F.concat(*rpn_pre_nms_proposals, dim=1)
            raw_rpn_scores = F.concat(*raw_rpn_scores, dim=1)
            raw_rpn_boxes = F.concat(*raw_rpn_boxes, dim=1)
        else:
            x = x[0]
            anchors = self.anchor_generator(x)
            x = self.conv1(x)
            raw_rpn_scores = self.score(x).transpose(axes=(0, 2, 3, 1)).reshape((0, -1, 1))
            rpn_scores = F.sigmoid(F.stop_gradient(raw_rpn_scores))
            raw_rpn_boxes = self.loc(x).transpose(axes=(0, 2, 3, 1)).reshape((0, -1, 4))
            rpn_boxes = F.stop_gradient(raw_rpn_boxes)
            rpn_pre_nms_proposals = self.region_proposer(
                anchors, rpn_scores, rpn_boxes, img)

        with autograd.pause():
            if self._per_level_nms and self._multi_level:
                # sort by scores
                tmp = F.contrib.box_nms(rpn_pre_nms_proposals, overlap_thresh=2.,
                                        topk=pre_nms + 1, coord_start=1, score_index=0, id_index=-1,
                                        force_suppress=True)
            else:
                # Non-maximum suppression
                tmp = F.contrib.box_nms(rpn_pre_nms_proposals, overlap_thresh=self._nms_thresh,
                                        topk=pre_nms, coord_start=1, score_index=0, id_index=-1,
                                        force_suppress=True)

            # slice post_nms number of boxes
            result = F.slice_axis(tmp, axis=1, begin=0, end=post_nms)
            rpn_scores = F.slice_axis(result, axis=-1, begin=0, end=1)
            rpn_boxes = F.slice_axis(result, axis=-1, begin=1, end=None)

        if autograd.is_training():
            # return raw predictions as well in training for bp
            return rpn_scores, rpn_boxes, raw_rpn_scores, raw_rpn_boxes, anchors
        return rpn_scores, rpn_boxes

    def get_test_post_nms(self):
        return self._test_post_nms


class RPNHead(gluon.HybridBlock):
    r"""Region Proposal Network Head.

    Parameters
    ----------
    channels : int
        Channel number used in convolutional layers.
    anchor_depth : int
        Each FPN stage have one scale and three ratios,
        we can compute anchor_depth = len(scale) \times len(ratios)
    """

    def __init__(self, channels, anchor_depth, **kwargs):
        super(RPNHead, self).__init__(**kwargs)
        weight_initializer = mx.init.Normal(0.01)
        with self.name_scope():
            self.conv1 = nn.HybridSequential()
            self.conv1.add(nn.Conv2D(channels, 3, 1, 1, weight_initializer=weight_initializer),
                           nn.Activation('relu'))
            # use sigmoid instead of softmax, reduce channel numbers
            # Note : that is to say, if use softmax here,
            # then the self.score will anchor_depth*2 output channel
            self.score = nn.Conv2D(anchor_depth, 1, 1, 0, weight_initializer=weight_initializer)
            self.loc = nn.Conv2D(anchor_depth * 4, 1, 1, 0, weight_initializer=weight_initializer)

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, x):
        """Forward RPN Head.

        This HybridBlock will generate predicted values for cls and box.

        Parameters
        ----------
        x : mxnet.nd.NDArray or mxnet.symbol
            Feature tensor. With (1, C, H, W) shape

        Returns
        -------
        (rpn_scores, rpn_boxes, raw_rpn_scores, raw_rpn_boxes)
            Returns predicted scores and regions.

        """
        # 3x3 conv with relu activation
        x = self.conv1(x)
        # (1, C, H, W)->(1, 9, H, W)->(1, H, W, 9)->(1, H*W*9, 1)
        raw_rpn_scores = self.score(x).transpose(axes=(0, 2, 3, 1)).reshape((0, -1, 1))
        # (1, H*W*9, 1)
        rpn_scores = F.sigmoid(F.stop_gradient(raw_rpn_scores))
        # (1, C, H, W)->(1, 36, H, W)->(1, H, W, 36)->(1, H*W*9, 4)
        raw_rpn_boxes = self.loc(x).transpose(axes=(0, 2, 3, 1)).reshape((0, -1, 4))
        # (1, H*W*9, 1)
        rpn_boxes = F.stop_gradient(raw_rpn_boxes)
        # return raw predictions as well in training for bp
        return rpn_scores, rpn_boxes, raw_rpn_scores, raw_rpn_boxes



"""Default setting in training/testing for directpose"""
from yacs.config import CfgNode as CN


_C = CN()

# ---------------------------------------------------------------------------- #
# Distributed DataParallel setting: DDP_CONFIG
# ---------------------------------------------------------------------------- #

_C.DDP_CONFIG = CN(new_allowed=False)
# Number of nodes for distributed training
_C.DDP_CONFIG.WORLD_SIZE = 1
# Node rank for distributed training
_C.DDP_CONFIG.WORLD_RANK = 0
# Number of GPUs to use
_C.DDP_CONFIG.GPU_WORLD_SIZE = 8
# GPU rank for distributed training
_C.DDP_CONFIG.GPU_WORLD_RANK = 0
# Master node
_C.DDP_CONFIG.DIST_URL = 'tcp://127.0.0.1:10001'
# A list of IP addresses for each node
_C.DDP_CONFIG.WOLRD_URLS = ['127.0.0.1']
# Whether to turn on automatic ranking match.
_C.DDP_CONFIG.AUTO_RANK_MATCH = True
# distributed backend
_C.DDP_CONFIG.DIST_BACKEND = 'nccl'
# Current GPU id.
_C.DDP_CONFIG.GPU = 0
# Whether to use distributed training or simply use dataparallel.
_C.DDP_CONFIG.DISTRIBUTED = True

# ---------------------------------------------------------------------------- #
# Standard training/testing setting: CONFIG
# ---------------------------------------------------------------------------- #

_C.CONFIG = CN(new_allowed=True)

_C.CONFIG.TRAIN = CN(new_allowed=True)
# Per GPU mini-batch size.
_C.CONFIG.TRAIN.BATCH_SIZE = 8
# Base learning rate.
_C.CONFIG.TRAIN.LR = 0.01
# Momentum.
_C.CONFIG.TRAIN.MOMENTUM = 0.9
# L2 regularization.
_C.CONFIG.TRAIN.W_DECAY = 1e-4
# The weight decay that's applied to parameters of normalization layers
_C.CONFIG.TRAIN.W_DECAY_NORM = 0.0
# Learning rate policy
_C.CONFIG.TRAIN.LR_POLICY: 'Cosine'
# Steps for multistep learning rate policy
_C.CONFIG.TRAIN.LR_MILESTONE = [40, 80]
# Exponential decay factor.
_C.CONFIG.TRAIN.STEP = 0.1
# Use warm up scheduler or not.
_C.CONFIG.TRAIN.USE_WARMUP: False
# Gradually warm up the TRAIN.LR over this number of epochs.
_C.CONFIG.TRAIN.WARMUP_EPOCHS: 34
# The LR scheduler name for the iteration based warm up, can be (WarmupMultiStepLR, WarmupCosineLR, WarmupLinearLR)
_C.CONFIG.TRAIN.ITER_LR_SCHEDULER_NAME = "WarmupMultiStepLR"
# Maximal number of epochs.
_C.CONFIG.TRAIN.ITER_NUM = 40000
# The iteration based number to decrease learning rate by SETP, it's similar but different from the epoch based LR_MILESTONE
_C.CONFIG.TRAIN.ITER_LR_STEPS = (30000,)
# Warm up reductio factor for iter based method
_C.CONFIG.TRAIN.ITER_BASED_WARMUP_FACTOR = 1.0 / 1000
# The maximum iterations for the warm up period
_C.CONFIG.TRAIN.ITER_BASED_WARMUP_ITERS = 1000
# The strategy for iteration based warm up
_C.CONFIG.TRAIN.ITER_BASED_WARMUP_METHOD = "linear"
# The start learning rate of the warm up.
_C.CONFIG.TRAIN.WARMUP_START_LR: 0.01
# The end learning rate of the warm up.
_C.CONFIG.TRAIN.WARMUP_END_LR: 0.1

# Gradient clipping
_C.CONFIG.TRAIN.CLIP_GRADIENTS = CN({"ENABLED": False})
# Type of gradient clipping, currently 2 values are supported:
# - "value": the absolute values of elements of each gradients are clipped
# - "norm": the norm of the gradient for each parameter is clipped thus
#   affecting all elements in the parameter
_C.CONFIG.TRAIN.CLIP_GRADIENTS.CLIP_TYPE = "value"
# Maximum absolute value used for clipping gradients
_C.CONFIG.TRAIN.CLIP_GRADIENTS.CLIP_VALUE = 1.0
# Floating point number p for L-p norm to be used with the "norm"
# gradient clipping type; for L-inf, please specify .inf
_C.CONFIG.TRAIN.CLIP_GRADIENTS.NORM_TYPE = 2.0


_C.CONFIG.DATA = CN(new_allowed=True)
_C.CONFIG.DATA.NUM_WORKERS = 0
# Switch for loading optional data
_C.CONFIG.DATA.LOAD_PROPOSALS = False
_C.CONFIG.DATA.MASK_ON = False
_C.CONFIG.DATA.MASK_FORMAT = "polygon"  # alternative: "bitmask"
_C.CONFIG.DATA.KEYPOINT_ON = False


# ----------------------------------------------------------------------------
# Registered dataset in registry(torch.data.registry.DatasetCatalog)
# ----------------------------------------------------------------------------
_C.CONFIG.DATA.DATASET = CN(new_allowed=True)
# List of the dataset names for training. Must be registered in DatasetCatalog
_C.CONFIG.DATA.DATASET.TRAIN = ()
# List of the pre-computed proposal files for training, which must be consistent
# with datasets listed in DATASETS.TRAIN.
_C.CONFIG.DATA.DATASET.PROPOSAL_FILES_TRAIN = ()
# Number of top scoring precomputed proposals to keep for training
_C.CONFIG.DATA.DATASET.PRECOMPUTED_PROPOSAL_TOPK_TRAIN = 2000
# List of the dataset names for testing. Must be registered in DatasetCatalog
_C.CONFIG.DATA.DATASET.VAL = ()
# List of the pre-computed proposal files for test, which must be consistent
# with datasets listed in DATASETS.TEST.
_C.CONFIG.DATA.DATASET.PROPOSAL_FILES_TEST = ()
# Number of top scoring precomputed proposals to keep for test
_C.CONFIG.DATA.DATASET.PRECOMPUTED_PROPOSAL_TOPK_TEST = 1000
_C.CONFIG.DATA.LOAD_PROPOSALS = False


# -----------------------------------------------------------------------------
# Detection Loader
# -----------------------------------------------------------------------------
_C.CONFIG.DATA.DETECTION = CN(new_allowed=True)
# If True, each batch should contain only images for which the aspect ratio
# is compatible. This groups portrait images together, and landscape images
# are not batched with portrait images.
_C.CONFIG.DATA.DETECTION.ASPECT_RATIO_GROUPING = True
# Options: TrainingSampler, RepeatFactorTrainingSampler
_C.CONFIG.DATA.DETECTION.SAMPLER_TRAIN = "TrainingSampler"
# Repeat threshold for RepeatFactorTrainingSampler
_C.CONFIG.DATA.DETECTION.REPEAT_THRESHOLD = 0.0
# Tf True, when working on datasets that have instance annotations, the
# training dataloader will filter out images without associated annotations
_C.CONFIG.DATA.DETECTION.FILTER_EMPTY_ANNOTATIONS = True
# Various Detection Size limitations
_C.CONFIG.DATA.DETECTION.MIN_SIZE_TRAIN_SAMPLING = "choice"
_C.CONFIG.DATA.DETECTION.MIN_SIZE_TRAIN = (800,)
_C.CONFIG.DATA.DETECTION.MAX_SIZE_TRAIN = 1333
_C.CONFIG.DATA.DETECTION.MIN_SIZE_TEST = 800
_C.CONFIG.DATA.DETECTION.MAX_SIZE_TEST = 1333
# `True` if cropping is used for data augmentation during training
_C.CONFIG.DATA.CROP = CN({"ENABLED": False}, new_allowed=True)
# Cropping type:
# - "relative" crop (H * CROP.SIZE[0], W * CROP.SIZE[1]) part of an input of size (H, W)
# - "relative_range" uniformly sample relative crop size from between [CROP.SIZE[0], [CROP.SIZE[1]].
#   and  [1, 1] and use it as in "relative" scenario.
# - "absolute" crop part of an input with absolute size: (CROP.SIZE[0], CROP.SIZE[1]).
# - "absolute_range", for an input of size (H, W), uniformly sample H_crop in
#   [CROP.SIZE[0], min(H, CROP.SIZE[1])] and W_crop in [CROP.SIZE[0], min(W, CROP.SIZE[1])]
_C.CONFIG.DATA.CROP.TYPE = "relative_range"
# Size of crop in range (0, 1] if CROP.TYPE is "relative" or "relative_range" and in number of
# pixels if CROP.TYPE is "absolute"
_C.CONFIG.DATA.CROP.SIZE = [0.9, 0.9]
_C.CONFIG.DATA.CROP.CROP_INSTANCE = True

# ------------------------------------------------------------------------------
# Model
# ------------------------------------------------------------------------------

_C.CONFIG.MODEL = CN(new_allowed=True)
# Model architecture. You can find available models in the model zoo.
_C.CONFIG.MODEL.NAME = ''
# Whether to load a checkpoint file. If True, please set the following PRETRAINED_PATH.
_C.CONFIG.MODEL.LOAD = False
# Path (a file path, or URL) to a checkpoint file to be loaded to the model.
_C.CONFIG.MODEL.PRETRAINED_PATH = ''
# Whether to use the trained weights in the model zoo.
_C.CONFIG.MODEL.PRETRAINED = False
# Whether to use pretrained backbone network. Usually this is set to True.
_C.CONFIG.MODEL.PRETRAINED_BASE = True
# BN options
_C.CONFIG.MODEL.BN_EVAL = False
_C.CONFIG.MODEL.PARTIAL_BN = False
_C.CONFIG.MODEL.BN_FROZEN = False
_C.CONFIG.MODEL.USE_AFFINE = False
# TVM trace
_C.CONFIG.MODEL.TVM_MODE = False

# BACKBONE
_C.CONFIG.MODEL.BACKBONE = CN(new_allowed=True)
_C.CONFIG.MODEL.BACKBONE.NAME = "build_directpose_resnet_fpn_backbone"
_C.CONFIG.MODEL.BACKBONE.ANTI_ALIAS = True
_C.CONFIG.MODEL.BACKBONE.FREEZE_AT = 0
_C.CONFIG.MODEL.BACKBONE.LOAD_URL = ''


_C.CONFIG.LOG = CN(new_allowed=True)
# Base directory where all output files are written
_C.CONFIG.LOG.BASE_PATH = ''
# Pre-defined name for each experiment.
# If set to 'use_time', the start time will be appended to the directory name.
_C.CONFIG.LOG.EXP_NAME = 'use_time'
# Directory where training logs are written
_C.CONFIG.LOG.LOG_DIR = 'tb_log'
# Directory where checkpoints are written
_C.CONFIG.LOG.SAVE_DIR = 'checkpoints'
# Directory where testing logs are written
_C.CONFIG.LOG.EVAL_DIR = ''
# Save a checkpoint after every this number of epochs
_C.CONFIG.LOG.SAVE_FREQ = 1
# Display the training log after every this number of iterations
_C.CONFIG.LOG.DISPLAY_FREQ = 1
# LOG LEVEL, choices are (CRITICAL, ERROR, WARNING, INFO, DEBUG)
_C.CONFIG.LOG.LEVEL = 'DEBUG'
# LOG output file, ends with .txt or .log
_C.CONFIG.LOG.LOG_FILENAME = 'log.txt'

# RESNET
_C.CONFIG.MODEL.RESNETS = CN(new_allowed=True)
_C.CONFIG.MODEL.RESNETS.OUT_FEATURES = ["res3", "res4", "res5"]
_C.CONFIG.MODEL.RESNETS.DEPTH = 50
_C.CONFIG.MODEL.RESNETS.NORM = "SyncBN"

# ---------------------------------------------------------------------------- #
# DLA backbone
# ---------------------------------------------------------------------------- #

_C.CONFIG.MODEL.DLA = CN()
_C.CONFIG.MODEL.DLA.CONV_BODY = "DLA34"
_C.CONFIG.MODEL.DLA.OUT_FEATURES = ["stage2", "stage3", "stage4", "stage5"]

# Options: FrozenBN, GN, "SyncBN", "BN"
_C.CONFIG.MODEL.DLA.NORM = "FrozenBN"

# ---------------------------------------------------------------------------- #
# FPN options
# ---------------------------------------------------------------------------- #
_C.CONFIG.MODEL.FPN = CN()
# Names of the input feature maps to be used by FPN
# They must have contiguous power of 2 strides
# e.g., ["res2", "res3", "res4", "res5"]
_C.CONFIG.MODEL.FPN.IN_FEATURES = []
_C.CONFIG.MODEL.FPN.OUT_CHANNELS = 256

# Options: "" (no norm), "GN"
_C.CONFIG.MODEL.FPN.NORM = ""

# Types for fusing the FPN top-down and lateral features. Can be either "sum" or "avg"
_C.CONFIG.MODEL.FPN.FUSE_TYPE = "sum"

# ---------------------------------------------------------------------------- #
# Keypoint Head
# ---------------------------------------------------------------------------- #
_C.CONFIG.MODEL.ROI_KEYPOINT_HEAD = CN()
_C.CONFIG.MODEL.ROI_KEYPOINT_HEAD.NUM_KEYPOINTS = 17  # 17 is the number of keypoints in COCO.
# Images with too few (or no) keypoints are excluded from training.
_C.CONFIG.MODEL.ROI_KEYPOINT_HEAD.MIN_KEYPOINTS_PER_IMAGE = 1

# ---------------------------------------------------------------------------- #
# DIRECTPOSE Head
# ---------------------------------------------------------------------------- #
_C.CONFIG.MODEL.DIRECTPOSE = CN()

# This is the number of foreground classes.
_C.CONFIG.MODEL.DIRECTPOSE.NUM_CLASSES = 1
_C.CONFIG.MODEL.DIRECTPOSE.NUM_KPTS = 17
_C.CONFIG.MODEL.DIRECTPOSE.IN_FEATURES = ["p3", "p4", "p5", "p6", "p7"]
_C.CONFIG.MODEL.DIRECTPOSE.FPN_STRIDES = [8, 16, 32, 64, 128]
_C.CONFIG.MODEL.DIRECTPOSE.PRIOR_PROB = 0.01
_C.CONFIG.MODEL.DIRECTPOSE.INFERENCE_TH_TRAIN = 0.05
_C.CONFIG.MODEL.DIRECTPOSE.INFERENCE_TH_TEST = 0.05
_C.CONFIG.MODEL.DIRECTPOSE.NMS_TH = 0.6
_C.CONFIG.MODEL.DIRECTPOSE.PRE_NMS_TOPK_TRAIN = 1000
_C.CONFIG.MODEL.DIRECTPOSE.PRE_NMS_TOPK_TEST = 1000
_C.CONFIG.MODEL.DIRECTPOSE.POST_NMS_TOPK_TRAIN = 100
_C.CONFIG.MODEL.DIRECTPOSE.POST_NMS_TOPK_TEST = 100
_C.CONFIG.MODEL.DIRECTPOSE.TOP_LEVELS = 2
_C.CONFIG.MODEL.DIRECTPOSE.NORM = "GN"  # Support GN or none
_C.CONFIG.MODEL.DIRECTPOSE.USE_SCALE = True

# Multiply centerness before threshold
# This will affect the final performance by about 0.05 AP but save some time
_C.CONFIG.MODEL.DIRECTPOSE.THRESH_WITH_CTR = False

# Focal loss parameters for i) Classification and ii) When using Binary Label as heatmap
_C.CONFIG.MODEL.DIRECTPOSE.LOSS_ALPHA = 0.25
_C.CONFIG.MODEL.DIRECTPOSE.LOSS_GAMMA = 2.0
_C.CONFIG.MODEL.DIRECTPOSE.SIZES_OF_INTEREST = [64, 128, 256, 512]
_C.CONFIG.MODEL.DIRECTPOSE.USE_RELU = True
_C.CONFIG.MODEL.DIRECTPOSE.USE_DEFORMABLE = False

# the number of convolutions used in the cls and bbox tower
_C.CONFIG.MODEL.DIRECTPOSE.NUM_CLS_CONVS = 4
_C.CONFIG.MODEL.DIRECTPOSE.NUM_BOX_CONVS = 4
_C.CONFIG.MODEL.DIRECTPOSE.NUM_KPT_CONVS = 4
_C.CONFIG.MODEL.DIRECTPOSE.NUM_HMS_CONVS = 2
_C.CONFIG.MODEL.DIRECTPOSE.NUM_SHARE_CONVS = 0
_C.CONFIG.MODEL.DIRECTPOSE.CENTER_SAMPLE = True
_C.CONFIG.MODEL.DIRECTPOSE.POS_RADIUS = 1.5
_C.CONFIG.MODEL.DIRECTPOSE.LOC_LOSS_TYPE = 'giou'

_C.CONFIG.MODEL.DIRECTPOSE.ENABLE_HM_BRANCH = True
_C.CONFIG.MODEL.DIRECTPOSE.HM_OFFSET = False
_C.CONFIG.MODEL.DIRECTPOSE.HM_CHANNELS = 128
_C.CONFIG.MODEL.DIRECTPOSE.HM_LOSS_TYPE = 'focal'
_C.CONFIG.MODEL.DIRECTPOSE.HM_LOSS_WEIGHT = 4.0
_C.CONFIG.MODEL.DIRECTPOSE.HM_TYPE = 'Gaussian'
_C.CONFIG.MODEL.DIRECTPOSE.HM_MSELOSS_BG_WEIGHT = 1.0      # When using MSE Loss on Gaussian heatmap
_C.CONFIG.MODEL.DIRECTPOSE.HM_MSELOSS_WEIGHT = 1.0
_C.CONFIG.MODEL.DIRECTPOSE.HM_FOCALLOSS_ALPHA = 2          # When using Focal Loss on Gaussian heatmap
_C.CONFIG.MODEL.DIRECTPOSE.HM_FOCALLOSS_BETA = 4
_C.CONFIG.MODEL.DIRECTPOSE.REFINE_KPT = False

_C.CONFIG.MODEL.DIRECTPOSE.KPT_VIS = False
_C.CONFIG.MODEL.DIRECTPOSE.LOSS_ON_LOCATOR = False
_C.CONFIG.MODEL.DIRECTPOSE.KPT_L1_beta = 1/9

_C.CONFIG.MODEL.DIRECTPOSE.ENABLE_BBOX_BRANCH = False
_C.CONFIG.MODEL.DIRECTPOSE.YIELD_PROPOSAL = False
_C.CONFIG.MODEL.DIRECTPOSE.SAMPLE_FEATURE = 'lower'
_C.CONFIG.MODEL.DIRECTPOSE.KPALIGN_GROUPS = 9
_C.CONFIG.MODEL.DIRECTPOSE.SEPERATE_CONV_FEATURE = True
_C.CONFIG.MODEL.DIRECTPOSE.SEPERATE_CONV_CHANNEL = 64

_C.CONFIG.MODEL.DIRECTPOSE.CLOSEKPT_NMS = False
_C.CONFIG.MODEL.DIRECTPOSE.CENTER_BRANCH = 'cls'
